""" Unit tests for pyart's retrieve/kdp_proc.py module. """

import numpy as np

from pyart.retrieve import kdp_proc
from pyart.filters import GateFilter
from pyart.testing import sample_objects
from pyart.config import get_field_name


def test_kdp_maesaka_linear_psidp(slope=0.002, maxiter=100):
    radar = _make_linear_psidp_radar(slope=slope)
    kdp_dict, phidpf_dict, phidpr_dict = kdp_proc.kdp_maesaka(
        radar, maxiter=maxiter, check_outliers=False)

    assert np.allclose(np.diff(kdp_dict['data'][0]), 0.0, atol=0.1)
    assert np.allclose(kdp_dict['data'], 1000.0 * slope / 2.0, atol=0.1)

    return


def test_kdp_maesaka_all_excluded(first_guess=0.01, maxiter=100):
    radar = _make_linear_psidp_radar()
    gatefilter = GateFilter(radar)
    gatefilter.exclude_all()
    kdp_dict, phidpf_dict, phidpr_dict = kdp_proc.kdp_maesaka(
        radar, gatefilter=gatefilter, first_guess=first_guess, maxiter=maxiter,
        check_outliers=False)

    assert np.allclose(kdp_dict['data'][0], 0.0, atol=first_guess)

    return


def _make_linear_psidp_radar(slope=0.002):
    """
    Create single-ray radar with linear differential phase profile with
    specified slope.

    Parameters
    ----------
    slope : float, optional
        Slope of differential phase profile in deg/m. Radar range gates cover
        0-1000 m, inclusive, with 10 m gate spacings.

    Returns
    -------
    radar : Radar
        Radar with linear differential phase profile in deg.

    """
    radar = sample_objects.make_empty_ppi_radar(101, 1, 1)
    psidp_dict = {
        'data': np.atleast_2d(np.linspace(0.0, slope * 1000.0, radar.ngates))
        }
    radar.add_field(get_field_name('differential_phase'), psidp_dict)

    return radar



"""
Retrieval of VADs from a radar object.

"""

import numpy as np

from ..config import get_field_name
from ..core import HorizontalWindProfile



def velocity_azimuth_display(radar, vel_field=None, z_want=None,
                             gatefilter=None):
    """
    Velocity azimuth display.

    Creates a VAD object containing U Wind, V Wind and height that
    can then be used to plot and produce the velocity azimuth display.

    Parameters
    ----------
    radar : Radar
        Radar object used.
    vel_field : string, optional
        Velocity field to use for VAD calculation.
    z_want : array, optional
        Heights for where to sample vads from.
        None will result in np.linespace(0, 10000, 100).
    gatefilter : GateFilter, optional
        A GateFilter indicating radar gates that should be excluded
        from the import vad calculation.

    Returns
    -------
    vad : HorizontalWindProfile
	A velocity azimuth display object containing height, speed, direction,
        u_wind, v_wind from a radar object.

    References
    ----------
    Michelson, D. B., Andersson, T., Koistinen, J., Collier, C. G., Riedl, J.,
    Szturc, J., Gjertsen, U., Nielsen, A. and Overgaard, S. (2000) BALTEX Radar
    Data Centre Products and their Methodologies. In SMHI Reports. Meteorology
    and Climatology. Swedish Meteorological and Hydrological Institute,
    Norrkoping.

    """
    speeds = []
    angles = []
    heights = []

    # Pulling z data from radar
    z_gate_data = radar.gate_z['data']

    # Setting parameters
    if z_want is None:
        z_want = np.linspace(0, 1000, 100)

    # Parse field parameters
    if vel_field is None:
        radar.check_field_exists('velocity')
        vel_field = get_field_name('velocity')

    # Selecting what velocity data to use based on gatefilter
    if gatefilter is not None:
        velocities = np.ma.masked_where(
            gatefilter.gate_excluded,
            radar.fields[vel_field]['data'])
    else:
        velocities = radar.fields[vel_field]['data']

    # Getting radar sweep index values
    for i in range(len(radar.sweep_start_ray_index['data'])):
        index_start = radar.sweep_start_ray_index['data'][i]
        index_end = radar.sweep_end_ray_index['data'][i]
        if not (index_end - index_start) % 2 == 0:
            index_end = index_end - 1

        used_velocities = velocities[index_start:index_end]
        azimuth = radar.azimuth['data'][index_start:index_end]
        elevation = radar.fixed_angle['data'][i]

        # Calculating speed and angle
        speed, angle = _vad_calculation(
            used_velocities, azimuth, elevation)

        print('max height', z_gate_data[index_start, :].max(),
              'meters')

        # Filling empty arrays with data
        speeds.append(speed)
        angles.append(angle)
        heights.append(z_gate_data[index_start, :])

    # Combining arrays and sorting
    speed_array = np.concatenate(speeds)
    angle_array = np.concatenate(angles)
    height_array = np.concatenate(heights)
    arg_order = height_array.argsort()
    speed_ordered = speed_array[arg_order]
    height_ordered = height_array[arg_order]
    angle_ordered = angle_array[arg_order]

    # Calculating U and V wind
    u_ordered, v_ordered = _sd_to_uv(speed_ordered, angle_ordered)
    u_mean = _interval_mean(u_ordered, height_ordered, z_want)
    v_mean = _interval_mean(v_ordered, height_ordered, z_want)
    vad = HorizontalWindProfile.from_u_and_v(
        z_want, u_mean, v_mean)
    return vad


def _vad_calculation(velocity_field, azimuth, elevation):
    """ Calculates VAD for a scan, returns speed and angle
    outdic = vad_algorithm(velocity_field, azimuth, elevation)
    velocity_field is a 2D array, azimuth is a 1D array,
    elevation is a number. All in degrees, m outdic contains
    speed and angle. """

    # Creating array with radar velocity data
    nrays, nbins = velocity_field.shape
    nrays2 = nrays // 2
    velocity_count = np.ma.empty((nrays2, nbins, 2))
    velocity_count[:, :, 0] = velocity_field[0:nrays2, :]
    velocity_count[:, :, 1] = velocity_field[nrays2:, :]

    # Converting from degress to radians
    sinaz = np.sin(np.deg2rad(azimuth))
    cosaz = np.cos(np.deg2rad(azimuth))

    # Masking array and testing for nan values
    sumv = np.ma.sum(velocity_count, 2)
    vals = np.isnan(sumv)
    vals2 = np.vstack((vals, vals))

    # Summing non-nan data and creating new array with summed data
    count = np.sum(np.isnan(sumv) == False, 0)
    count = np.float64(count)
    u_m = np.array([np.nansum(sumv, 0) // (2 * count)])

    # Creating 0 value arrays
    cminusu_mcos = np.zeros((nrays, nbins))
    cminusu_msin = np.zeros((nrays, nbins))
    sincos = np.zeros((nrays, nbins))
    sin2 = np.zeros((nrays, nbins))
    cos2 = np.zeros((nrays, nbins))

    # Summing all sin and cos and setting select entires to nan
    for i in range(nbins):
        cminusu_mcos[:, i] = cosaz * (velocity_field[:, i] - u_m[:, i])
        cminusu_msin[:, i] = sinaz * (velocity_field[:, i] - u_m[:, i])
        sincos[:, i] = sinaz * cosaz
        sin2[:, i] = sinaz**2
        cos2[:, i] = cosaz**2

    cminusu_mcos[vals2] = np.nan
    cminusu_msin[vals2] = np.nan
    sincos[vals2] = np.nan
    sin2[vals2] = np.nan
    cos2[vals2] = np.nan
    sumcminu_mcos = np.nansum(cminusu_mcos, 0)
    sumcminu_msin = np.nansum(cminusu_msin, 0)
    sumsincos = np.nansum(sincos, 0)
    sumsin2 = np.nansum(sin2, 0)
    sumcos2 = np.nansum(cos2, 0)

    # Calculating speed and angle values
    b_value = (sumcminu_mcos - (sumsincos*sumcminu_msin / sumsin2)) / (
        sumcos2 - (sumsincos**2) / sumsin2)
    a_value = (sumcminu_msin - b_value*sumsincos) / sumsin2
    speed = np.sqrt(a_value**2 + b_value**2) / np.cos(
        np.deg2rad(elevation))
    angle = np.arctan2(a_value, b_value)
    return speed, angle


def _interval_mean(data, current_z, wanted_z):
    """ Find the mean of data indexed by current_z
        at wanted_z on intervals wanted_z+/- delta
        wanted_z. """
    delta = wanted_z[1] - wanted_z[0]
    pos_lower = [np.argsort((current_z - (
        wanted_z[i] - delta / 2.0))**2)[0]
                 for i in range(len(wanted_z))]
    pos_upper = [np.argsort((current_z - (
        wanted_z[i] + delta / 2.0))**2)[0]
                 for i in range(len(wanted_z))]
    mean_values = np.array([data[pos_lower[i]:pos_upper[i]].mean()
                            for i in range(len(pos_upper))])
    return mean_values


def _sd_to_uv(speed, direction):
    """ Takes speed and direction to create u_mean and v_mean. """
    return (np.sin(direction) * speed), (np.cos(direction) * speed)




import pyart
import argparse

if __name__ == '__main__':

    # parse the arguments
    parser = argparse.ArgumentParser(
        description='Convert a radar file to CF/Radial format.')
    parser.add_argument('infile', type=str, help='radar file to convert')
    parser.add_argument('outfile', type=str, help='file to save file to')

    igroup = parser.add_argument_group(
        title='ingest method, optional',
        description=('The method of file ingest can be specified. '
                     'If no ingest is specified, the format of the file will '
                     'be used to determine the best ingest method. '
                     'Specify only one of the following:'))

    igroup.add_argument('--sigmet', action='store_true',
                        help='Sigmet/IRIS ingest')
    igroup.add_argument('--mdv', action='store_true', help='MDV ingest')
    igroup.add_argument('--cfradial', action='store_true',
                        help='CF/Radial ingest')
    igroup.add_argument('--rsl', action='store_true',
                        help='RSL ingest')
    igroup.add_argument('--nexrad_archive', action='store_true',
                        help='NEXRAD level 2 archive ingest')
    igroup.add_argument('--nexrad_cdm', action='store_true',
                        help='NEXRAD level 2 CDM ingest')

    fgroup = parser.add_mutually_exclusive_group()
    fgroup.add_argument('--netcdf3_classic', action='store_const',
                        const='NETCDF3_CLASSIC', dest='netcdf_format',
                        help='create a NetCDF3 classic formatted file')
    fgroup.add_argument('--netcdf3_64bit', action='store_const',
                        const='NETCDF3_64BIT', dest='netcdf_format',
                        help='create a NetCDF3 64-bit offset formatted file')
    fgroup.add_argument('--netcdf4', action='store_const',
                        const='NETCDF4', dest='netcdf_format',
                        help='create a NetCDF4 formatted file (default)')
    fgroup.add_argument('--netcdf4_classic', action='store_const',
                        const='NETCDF4_CLASSIC', dest='netcdf_format',
                        help='create a NetCDF4 classic formatted file')
    parser.set_defaults(netcdf_format='NETCDF4')

    parser.add_argument('-v', '--version', action='version',
                        version='Py-ART version %s' % (pyart.__version__))
    args = parser.parse_args()

    # read in the file
    if args.sigmet:
        radar = pyart.io.read_sigmet(args.infile)
    elif args.mdv:
        radar = pyart.io.read_mdv(args.infile)
    elif args.cfradial:
        radar = pyart.io.read_cfradial(args.infile)
    elif args.rsl:
        radar = pyart.io.read_rsl(args.infile)
    elif args.nexrad_archive:
        radar = pyart.io.read_nexrad_archive(args.infile)
    elif args.nexrad_cdm:
        radar = pyart.io.read_nexrad_cdm(args.infile)
    else:
        radar = pyart.io.read(args.infile)

    # convert to CF/Radial
    pyart.io.write_cfradial(args.outfile, radar, format=args.netcdf_format)



"""
Routines for reading files from the X-band radar from the PATTERN_ project.

.. _PATTERN: http://www.mi.uni-hamburg.de/PATTERN-Pre.6763.0.html

"""

import datetime

import numpy as np
import netCDF4

from ..config import FileMetadata
from ..io.common import make_time_unit_str, _test_arguments
from ..core.radar import Radar


def read_pattern(filename, **kwargs):
    """
    Read a netCDF file from a PATTERN project X-band radar.

    Parameters
    ----------
    filename : str
        Name of netCDF file to read data from.

    Returns
    -------
    radar : Radar
        Radar object.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata('pattern')

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables

    # general parameters
    nrays = ncvars['Azimuth'].shape[0]
    scan_type = 'ppi'

    # time
    # interpolate between the first and last timestamps in the Time variable
    time = filemetadata('time')
    nctime = ncvars['Time']
    time['units'] = make_time_unit_str(
        datetime.datetime.utcfromtimestamp(nctime[0]))
    time['data'] = np.linspace(0, nctime[-1] - nctime[0], nrays)

    # range
    _range = filemetadata('range')
    _range['data'] = ncvars['Distance'][:]
    _range['meters_to_center_of_first_gate'] = _range['data'][0]
    # assuming the distance between all gates is constant, may not
    # always be true.
    _range['meters_between_gates'] = (_range['data'][1] - _range['data'][0])

    # fields
    # files contain a single corrected reflectivity field
    fields = {}
    field_name = filemetadata.get_field_name('corrected_reflectivity')
    field_dic = filemetadata(field_name)
    field_dic['_FillValue'] = ncvars['Corrected_Reflectivity']._FillValue
    field_dic['data'] = ncvars['Corrected_Reflectivity'][:]
    fields[field_name] = field_dic

    # metadata
    metadata = filemetadata('metadata')
    for k in ['institution', 'title', 'used_algorithms']:
        if k in ncobj.ncattrs():
            metadata[k] = ncobj.getncattr(k)

    # latitude, longitude, altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')
    latitude['data'] = np.array([ncobj.latitude[:-1]], dtype='float64')
    longitude['data'] = np.array([ncobj.longitude[:-1]], dtype='float64')
    altitude['data'] = np.array([ncobj.elevation], dtype='float64')

    # sweep parameters
    # sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
    # sweep_end_ray_index
    sweep_number = filemetadata('sweep_number')
    sweep_mode = filemetadata('sweep_mode')
    fixed_angle = filemetadata('fixed_angle')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')

    sweep_number['data'] = np.arange(1, dtype='int32')
    sweep_mode['data'] = np.array(1 * ['azimuth_surveillance'])
    fixed_angle['data'] = np.array([0], dtype='float32')
    sweep_start_ray_index['data'] = np.array([0], dtype='int32')
    sweep_end_ray_index['data'] = np.array([nrays-1], dtype='int32')

    # azimuth, elevation
    azimuth = filemetadata('azimuth')
    elevation = filemetadata('elevation')

    azimuth['data'] = ncvars['Azimuth'][:]
    elevation['data'] = np.array([0.], dtype='float32')

    # instrument parameters
    instrument_parameters = None

    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)



"""
Class for reading data from NEXRAD Level 3 files.

"""

# This file is part of the Py-ART, the Python ARM Radar Toolkit
# https://github.com/ARM-DOE/pyart

# Care has been taken to keep this file free from extraneous dependancies
# so that it can be used by other projects with no/minimal modification.

# Please feel free to use this file in other project provided the license
# below is followed.  Keeping the above comment lines would also be helpful
# to direct other back to the Py-ART project and the source of this file.


LICENSE = """
Copyright (c) 2013, UChicago Argonne, LLC
All rights reserved.

Copyright 2013 UChicago Argonne, LLC. This software was produced under U.S.
Government contract DE-AC02-06CH11357 for Argonne National Laboratory (ANL),
which is operated by UChicago Argonne, LLC for the U.S. Department of Energy.
The U.S. Government has rights to use, reproduce, and distribute this
software. NEITHER THE GOVERNMENT NOR UCHICAGO ARGONNE, LLC MAKES ANY
WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LIABILITY FOR THE USE OF THIS
SOFTWARE. If software is modified to produce derivative works, such modified
software should be clearly marked, so as not to confuse it with the version
available from ANL.

Additionally, redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following conditions
are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    * Neither the name of UChicago Argonne, LLC, Argonne National
      Laboratory, ANL, the U.S. Government, nor the names of its
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY UCHICAGO ARGONNE, LLC AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL UCHICAGO ARGONNE, LLC OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import bz2
from datetime import datetime, timedelta
import struct

import numpy as np


class NEXRADLevel3File(object):
    """
    A Class for accessing data in NEXRAD Level III (3) files.

    Attributes
    ----------
    text_header : dic
        File textual header.
    msg_header : dic
        Message header.
    prod_descr : dic
        Product description.
    symbology_header : dict
        Symbology header.
    packet_header : dict
        Radial data array packet header.
    radial_headers : list of dicts
        List of radials headers.
    raw_data : array
        Raw unscaled, unmasked data.
    data : array
        Scaled, masked radial data.
    _fh : file-like
        File like object from which data is read.

    """

    def __init__(self, filename):
        """ initalize the object. """
        # read the entire file into memory
        if hasattr(filename, 'read'):
            fhandle = filename
        else:
            fhandle = open(filename, 'rb')
        buf = fhandle.read()    # string buffer containing file data
        self._fh = fhandle

        # Text header
        # Format of Text header is SDUSXX KYYYY DDHHMM\r\r\nAAABBB\r\r\n
        # Sometime additional padding is present before the Text header
        record_padding = buf.find(b'SDUS')
        if record_padding == -1:
            raise ValueError('Not a valid NEXRAD Level 3 file.')
        self.text_header = buf[:30 + record_padding]
        bpos = 30 + record_padding      # current reading position in buffer

        # Read and decode 18 byte Message Header Block
        self.msg_header = _unpack_from_buf(buf, bpos, MESSAGE_HEADER)
        if self.msg_header['code'] not in SUPPORTED_PRODUCTS:
            code = self.msg_header['code']
            raise NotImplementedError(
                'Level3 product with code %i is not supported' % (code))
        bpos += 18

        # Read and decode 102 byte Product Description Block
        self.prod_descr = _unpack_from_buf(buf, bpos, PRODUCT_DESCRIPTION)
        bpos += 102

        # uncompressed symbology block if necessary
        if buf[bpos:bpos+2] == b'BZ':
            buf2 = bz2.decompress(buf[bpos:])
        else:
            buf2 = buf[bpos:]

        self._read_symbology_block(buf2)

    def close(self):
        """ Close the file. """
        self._fh.close()

    def _read_symbology_block(self, buf2):
        """ Read symbology block. """
        # Read and decode symbology header
        self.symbology_header = _unpack_from_buf(buf2, 0, SYMBOLOGY_HEADER)

        # Read radial packets
        packet_code = struct.unpack('>h', buf2[16:18])[0]
        assert packet_code in SUPPORTED_PACKET_CODES
        self.packet_header = _unpack_from_buf(buf2, 16, RADIAL_PACKET_HEADER)
        self.radial_headers = []
        nbins = self.packet_header['nbins']
        nradials = self.packet_header['nradials']
        nbytes = _unpack_from_buf(buf2, 30, RADIAL_HEADER)['nbytes']
        if packet_code == 16 and nbytes != nbins:
            nbins = nbytes  # sometimes these do not match, use nbytes
        self.raw_data = np.empty((nradials, nbins), dtype='uint8')
        pos = 30

        for radial in self.raw_data:
            radial_header = _unpack_from_buf(buf2, pos, RADIAL_HEADER)
            pos += 6
            if packet_code == 16:
                radial[:] = np.frombuffer(buf2[pos:pos+nbins], '>u1')
                pos += radial_header['nbytes']
            else:
                assert packet_code == AF1F
                # decode run length encoding
                rle_size = radial_header['nbytes'] * 2
                rle = np.frombuffer(buf2[pos:pos+rle_size], dtype='>u1')
                colors = np.bitwise_and(rle, 0b00001111)
                runs = np.bitwise_and(rle, 0b11110000) // 16
                radial[:] = np.repeat(colors, runs)
                pos += rle_size
            self.radial_headers.append(radial_header)

    def get_location(self):
        """ Return the latitude, longitude and height of the radar. """
        latitude = self.prod_descr['latitude'] * 0.001
        longitude = self.prod_descr['longitude'] * 0.001
        height = self.prod_descr['height']
        return latitude, longitude, height

    def get_azimuth(self):
        """ Return an array of starting azimuth angles in degrees. """
        azimuths = [d['angle_start'] for d in self.radial_headers]
        return np.array(azimuths, dtype='float32') * 0.1

    def get_range(self):
        """ Return an array of gate range spacing in meters. """
        nbins = self.raw_data.shape[1]
        first_bin = self.packet_header['first_bin']
        range_scale = (self.packet_header['range_scale'] *
                       PRODUCT_RANGE_RESOLUTION[self.msg_header['code']])
        return np.arange(nbins, dtype='float32') * range_scale + first_bin

    def get_elevation(self):
        """ Return the sweep elevation angle in degrees. """
        hw30 = self.prod_descr['halfwords_30']
        elevation = struct.unpack('>h', hw30)[0] * 0.1
        return elevation

    def get_volume_start_datetime(self):
        """ Return a datetime of the start of the radar volume. """
        return _datetime_from_mdate_mtime(self.prod_descr['vol_scan_date'],
                                          self.prod_descr['vol_scan_time'])

    def get_data(self):
        """ Return an masked array containing the field data. """
        msg_code = self.msg_header['code']
        threshold_data = self.prod_descr['threshold_data']

        if msg_code in _8_OR_16_LEVELS:
            mdata = self._get_data_8_or_16_levels()

        elif msg_code in [134]:
            mdata = self._get_data_msg_134()

        elif msg_code in [94, 99, 182, 186]:
            hw31, hw32 = np.frombuffer(threshold_data[:4], '>i2')
            data = (self.raw_data - 2) * (hw32/10.) + hw31/10.
            mdata = np.ma.array(data, mask=self.raw_data < 2)

        elif msg_code in [32]:
            hw31, hw32 = np.frombuffer(threshold_data[:4], '>i2')
            data = (self.raw_data) * (hw32/10.) + hw31/10.
            mdata = np.ma.array(data, mask=self.raw_data < 2)

        elif msg_code in [138]:
            hw31, hw32 = np.frombuffer(threshold_data[:4], '>i2')
            data = self.raw_data * (hw32/100.) + hw31/100.
            mdata = np.ma.array(data)

        elif msg_code in [159, 161, 163]:
            scale, offset = np.frombuffer(threshold_data[:8], '>f4')
            data = (self.raw_data - offset) / (scale)
            mdata = np.ma.array(data, mask=self.raw_data < 2)

        elif msg_code in [170, 172, 173, 174, 175]:
            # units are 0.01 inches
            scale, offset = np.frombuffer(threshold_data[:8], '>f4')
            data = (self.raw_data - offset) / (scale) * 0.01
            mdata = np.ma.array(data, mask=self.raw_data < 1)

        elif msg_code in [165, 177]:
            # Corresponds to classifications in table on page 3-37
            mdata = np.ma.masked_equal(self.raw_data, 0)

        elif msg_code in [135]:
            mdata = np.ma.array(self.raw_data - 2, mask=self.raw_data <= 1)
            mdata[self.raw_data >= 128] -= np.uint8(128)

        else:
            assert msg_code in [34]
            # There does not seem to be any discussion on what this product
            # contains.
            mdata = np.ma.array(self.raw_data.copy())

        return mdata.astype('float32')

    def _get_data_8_or_16_levels(self):
        """ Return a masked array for products with 8 or 16 data levels. """
        thresh = np.frombuffer(self.prod_descr['threshold_data'], '>B')
        flags = thresh[::2]
        values = thresh[1::2]

        sign = np.choose(np.bitwise_and(flags, 0x01), [1, -1])
        bad = np.bitwise_and(flags, 0x80) == 128
        scale = 1.
        if flags[0] & 2**5:
            scale = 1/20.
        if flags[0] & 2**4:
            scale = 1/10.

        data_levels = values * sign * scale
        data_levels[bad] = -999 # sentinal for bad data points

        data = np.choose(self.raw_data, data_levels)
        mdata = np.ma.masked_equal(data, -999)
        return mdata

    def _get_data_msg_134(self):
        """ Return a masked array for product with message code 134. """
        hw31, hw32, hw33, hw34, hw35 = np.frombuffer(
            self.prod_descr['threshold_data'][:10], '>i2')
        linear_scale = _int16_to_float16(hw31)
        linear_offset = _int16_to_float16(hw32)
        log_start = hw33
        log_scale = _int16_to_float16(hw34)
        log_offset = _int16_to_float16(hw35)
        # linear scale data
        data = np.zeros(self.raw_data.shape, dtype=np.float32)
        lin = self.raw_data < log_start
        data[lin] = ((self.raw_data[lin] - linear_offset) / (linear_scale))
        # log scale data
        log = self.raw_data >= log_start
        data[log] = np.exp((self.raw_data[log] - log_offset) / (log_scale))
        mdata = np.ma.masked_array(data, mask=self.raw_data < 2)
        return mdata


def _datetime_from_mdate_mtime(mdate, mtime):
    """ Returns a datetime for a given message date and time. """
    epoch = datetime.utcfromtimestamp(0)
    return epoch + timedelta(days=mdate - 1, seconds=mtime)


def _structure_size(structure):
    """ Find the size of a structure in bytes. """
    return struct.calcsize('>' + ''.join([i[1] for i in structure]))


def _unpack_from_buf(buf, pos, structure):
    """ Unpack a structure from a buffer. """
    size = _structure_size(structure)
    return _unpack_structure(buf[pos:pos + size], structure)


def _unpack_structure(string, structure):
    """ Unpack a structure from a string """
    fmt = '>' + ''.join([i[1] for i in structure]) # NEXRAD is big-endian
    lst = struct.unpack(fmt, string)
    return dict(zip([i[0] for i in structure], lst))


def nexrad_level3_message_code(filename):
    """ Return the message (product) code for a NEXRAD Level 3 file. """
    fhl = open(filename, 'r')
    buf = fhl.read(48)
    fhl.close()
    msg_header = _unpack_from_buf(buf, 30, MESSAGE_HEADER)
    return msg_header['code']


# NEXRAD Level III file structures, sizes, and static data
# The deails on these structures are documented in:
# "INTERFACE CONTROL DOCUMENT FOR THE RPG TO CLASS 1 USER" RPG Build 13.0
# Document Number 2620001T
# Tables and page number refer to those in this document.


def _int16_to_float16(val):
    """ Convert a 16 bit interger into a 16 bit float. """
    # NEXRAD Level III float16 format defined on page 3-33.
    # Differs from IEEE 768-2008 format so np.float16 cannot be used.
    sign = (val & 0b1000000000000000) / 0b1000000000000000
    exponent = (val & 0b0111110000000000) / 0b0000010000000000
    fraction = (val & 0b0000001111111111)
    if exponent == 0:
        return (-1)**sign * 2 * (0 + (fraction/2**10.))
    else:
        return (-1)**sign * 2**(exponent-16) * (1 + fraction/2**10.)


_8_OR_16_LEVELS = [19, 20, 25, 27, 28, 30, 56, 78, 79, 80, 169, 171, 181]

PRODUCT_RANGE_RESOLUTION = {
    19: 1.,     # 124 nm
    20: 2.,     # 248 nm
    25: 0.25,   # 32 nm
    27: 1.,
    28: 0.25,
    30: 1.,
    32: 1.,
    34: 1.,
    56: 1.,
    78: 1.,
    79: 1.,
    80: 1.,
    94: 1.,
    99: 0.25,
    134: 1000.,
    135: 1000.,
    138: 1.,
    159: 0.25,
    161: 0.25,
    163: 0.25,
    165: 0.25,
    169: 1.,
    170: 1.,
    171: 1.,
    172: 1.,
    173: 1.,
    174: 1.,
    175: 1.,
    177: 0.25,
    181: 150.,
    182: 150.,
    186: 300.,
}


# format of structure elements
# Figure E-1, page E-1
BYTE = 'B'      # not in table but used in Product Description
INT2 = 'h'
INT4 = 'i'
UINT4 = 'I'
REAL4 = 'f'

# 3.3.1 Graphic Product Messages

# Graphic Product Message: Message Header Block
# 18 bytes, 9 halfwords
# Figure 3-3, page 3-7.
MESSAGE_HEADER = (
    ('code', INT2),     # message code
    ('date', INT2),     # date of message, days since 1 Jan, 1970
    ('time', INT4),     # time of message, seconds since midnight
    ('length', INT4),   # length of message in bytes
    ('source', INT2),   # Source ID
    ('dest', INT2),     # Destination ID
    ('nblocks', INT2),  # Number of blocks in the message (inclusive)
)

# Graphic Product Message: Product Description Block
# Description: section 3.3.1.1, page 3-3
# 102 bytes, 51 halfwords (halfwords 10-60)
# Figure 3-6, pages 3-31 and 3-32
PRODUCT_DESCRIPTION = (
    ('divider', INT2),          # Delineate blocks, -1
    ('latitude', INT4),         # Latitude of radar, degrees, + for north
    ('longitude', INT4),        # Longitude of radar, degrees, + for east
    ('height', INT2),           # Height of radar, feet abouve mean sea level
    ('product_code', INT2),     # NEXRAD product code
    ('operational_mode', INT2),  # 0 = Maintenance, 1 = Clean Air, 2 = Precip
    ('vcp', INT2),              # Volume Coverage Pattern of scan strategy
    ('sequence_num', INT2),     # Sequence Number of the request.
    ('vol_scan_num', INT2),     # Volume Scan number, 1 to 80.
    ('vol_scan_date', INT2),    # Volume Scan start date, days since 1/1/1970
    ('vol_scan_time', INT4),    # Volume Scan start time, sec since midnight
    ('product_date', INT2),     # Product Generation Date, days since 1/1/1970
    ('product_time', INT4),     # Product Generation Time, sec since midnight
    ('halfwords_27_28', '4s'),  # Product dependent parameters 1 and 2
    ('elevation_num', INT2),    # Elevation number within volume scan
    ('halfwords_30', '2s'),     # Product dependent parameter 3
    ('threshold_data', '32s'),  # Data to determine threshold level values
    ('halfwords_47_53', '14s'),  # Product dependent parameters 4-10
    ('version', BYTE),          # Version, 0
    ('spot_blank', BYTE),       # 1 = Spot blank ON, 0 = Blanking OFF
    ('offet_symbology', INT4),  # halfword offset to Symbology block
    ('offset_graphic', INT4),   # halfword offset to Graphic block
    ('offset_tabular', INT4)    # halfword offset to Tabular block
)

# Graphic Product Message: Product Symbology Block
# Description
# 16 byte header
# Figure 3-6 (Sheet 8), pages 3-40

SYMBOLOGY_HEADER = (
    ('divider', INT2),          # Delineate blocks, -1
    ('id', INT2),               # Block ID, 1
    ('block_length', INT4),     # Length of block in bytes
    ('layers', INT2),           # Number of data layers
    ('layer_divider', INT2),    # Delineate data layers, -1
    ('layer_length', INT4)      # Length of data layer in bytes
    # Display data packets
)

# Digital Radial Data Array Packet - Packet Code 16 (Sheet 2)
# Figure 3-11c (Sheet 1 and 2), page 3-120.
# and
# Radial Data Packet - Packet Code AF1F
# Figure 3-10 (Sheet 1 and 2), page 3-113.
AF1F = -20705       # struct.unpack('>h', 'AF1F'.decode('hex'))
SUPPORTED_PACKET_CODES = [16, AF1F]       # elsewhere
RADIAL_PACKET_HEADER = (
    ('packet_code', INT2),      # Packet Code, Type 16
    ('first_bin', INT2),        # Location of first range bin.
    ('nbins', INT2),            # Number of range bins.
    ('i_sweep_center', INT2),   # I coordinate of center of sweep.
    ('j_sweep_center', INT2),   # J coordinate of center of sweep.
    ('range_scale', INT2),      # Range Scale factor
    ('nradials', INT2)          # Total number of radials in the product
)

RADIAL_HEADER = (
    ('nbytes', INT2),           # Number of bytes in the radial.
    ('angle_start', INT2),      # Starting angle at which data was collected.
    ('angle_delta', INT2)       # Delta angle from previous radial.
)

# A list of the NEXRAD Level 3 Product supported by this module taken
# from the "Message Code for Products" Table III pages 3-15 to 3-22
# All the supported products have a Radial Image Message format.
#   Code    # Product Name
#   -----   -----------------------
SUPPORTED_PRODUCTS = [
    19,     # Base Reflectivity
    20,     # Base Reflectivity
    25,     # Base Velocity
    27,     # Base Velocity
    28,     # Base Spectrum Width
    30,     # Base Spectrum Width
    32,     # Digital Hybrid Scan
    34,     # Clutter Filter Control
    56,     # Storm Relative Mean
            # Radial Velocity
    78,     # Surface Rainfall Accum.
            # (1 hr)
    79,     # Surface Rainfall Accum.
            # (3 hr)
    80,     # Storm Total Rainfall
            # Accumulation
    94,     # Base Reflectivity Data
            # Array
    99,     # Base Velocity Data
            # Array
    134,    # High Resolution VIL
    135,    # Enhanced Echo Tops
    138,    # Digital Storm Total
            # Precipitation
    159,    # Digital Differential
            # Reflectivity
    161,    # Digital Correlation
            # Coefficient
    163,    # Digital Specific
            # Differential Phase
    165,    # Digital Hydrometeor
            # Classification
    169,    # One Hour Accumulation
    170,    # Digital Accumulation
            # Array
    171,    # Storm Total
            # Accumulation
    172,    # Digital Storm Total
            # Accumulation
    173,    # Digital User-Selectable
            # Accumulation
    174,    # Digital One-Hour
            # Difference Accumulation
    175,    # Digital Storm Total
            # Difference Accumulation
    177,    # Hybrid Hydrometeor
            # Classification
    181,    # Base Reflectivity
    182,    # Base Velocity
    186,    # Base Reflectivity
]

# It should be possible to add support for these NEXRAD Level 3 products
# as they are of Radial Image message format.  No examples of these files
# could be found to test on so support does not exist yet in this module.
#    Code    # Product Name
#    -----   -----------------------
#    16,     # Base Reflectivity
#    17,     # Base Reflectivity
#    18,     # Base Reflectivity
#    21,     # Base Reflectivity
#    22,     # Base Velocity
#    23,     # Base Velocity
#    24,     # Base Velocity
#    26,     # Base Velocity
#    29,     # Base Spectrum Width
#    31,     # User Selectable Storm Total Precipitation Reflectivty
#    33,     # Hybrid Scan Reflectivty
#    55,     # Storm Relative Mean Radial Velocity
#    93,     # ITWS Digital Base Velocity
#    132,    # Clutter Likelihood Reflectivity
#    133,    # Clutter Likelihood Doppler
#    137,    # User Selectable Layer Composite Reflectivity
#    144,    # One-hour Snow Water Equivalent
#    145,    # One-hour Snow Depth
#    146,    # Storm Total Snow Water Equivalent
#    147,    # Storm Total Snow Depth
#    150,    # User Selectable Snow Water Equivalent
#    151,    # User Selectable Snow Depth
#    153,    # Super Resolution Reflectivity Data Array
#    154,    # Super Resolution Velocity Data Array
#    155,    # Super Resolution Spectrum Width Data Array
#    158,    # Differential Reflectivity
#    160,    # Correlation Coefficient
#    162,    # Specific Differential Phase
#    164,    # Hydrometeor Classification
#    194,    # Base Reflectivity Data Array (DoD Version)
#    195,    # Digital Reflectivity, DQA-Edited Data Array
#    199,    # Base Velocity Data Array (DoD Version)
#    183,    # TDWR Base Velocity
#    180,    # TDWR Base Reflectivity
#    185,    # TDWR Base Spectrum Width
#    187,    # TDWR Base Reflectivity

# No support for these NEXRAD Level 3 products is planned as they
# do not have a Radial Image message format.

#   Code    # Product Name             Message format
#   -----   -----------------------    ---------------
#   35,     # Composite Reflectivity  Raster Image/Nongeographic Alpha
#   36,     # Composite Reflectivity  Raster Image/Nongeographic Alpha
#   37,     # Composite Reflectivity  Raster Image/Nongeographic Alpha
#   38,     # Composite Reflectivity  Raster Image/Nongeographic Alpha
#   39,     # Spare
#   40,     # Spare
#   41,     # Echo Tops               Raster Image
#   42,     # Spare
#   43,     # Spare
#   44,     # Spare
#   45,     # Spare
#   46,     # Spare
#   47,     # Spare
#   48,     # VAD Wind Profile        Non-geographic Alphanumeric
#   49,     # Spare
#   50,     # Cross Section           Raster Image (Reflectivity)
#           # (Reflectivity)
#   51,     # Cross Section           Raster Image (Velocity)
#           # (Velocity)
#   52,     # Spare
#   53,     # Spare
#   54,     # Reserved
#   57,     # Vertically Integrated   Raster Image
#           # Liquid
#   58,     # Storm Tracking          Non-geographic Alpha
#           # Information
#   59,     # Hail Index              Non-geographic Alpha
#   60,     # Spare
#   61,     # Tornado Vortex          Geographic and Non-geographic
#           # Signature               Alphanumeric
#   62,     # Storm Structure         Alphanumeric
#   63,     # Layer Composite         Raster Image (Layer 1 Average)
#           # Reflectivity
#   64,     # Layer Composite         Raster Image (Layer 2 Average)
#           # Reflectivity
#   65,     # Layer Composite         Raster Image (Layer 1 Maximum)
#           # Reflectivity
#   66,     # Layer Composite         Raster Image (Layer 2 Maximum)
#           # Reflectivity
#   67,     # Layer Composite         Raster Image
#           # Reflectivity - AP Removed
#   68,     # Spare
#   69,     # Spare
#   70,     # Spare
#   71,     # Spare
#   72,     # Spare
#   73,     # User Alert Message      Alphanumeric
#   74,     # Radar Coded Message     Alphanumeric
#   75,     # Free Text Message       Alphanumeric
#   76,     # Reserved for internal PUP use.
#   81,     # Hourly Digital          Raster Image / Alphanumeric
#           # Precipitation Array
#   82,     # Supplemental            Alphanumeric
#           # Precipitation Data
#   83,     # Spare
#   84,     # Velocity Azimuth        Non-geographic Alphanumeric
#           # Display
#   85,     # Cross Section           Raster Image (Reflectivity)
#           # Reflectivity
#   86,     # Cross Section Velocity  Raster Image (Velocity)
#   87,     # Spare
#   88,     # Spare
#   89,     # Layer Composite         Raster Image - Layer 3 Average
#           # Reflectivity
#   90,     # Layer Composite         Raster Image - Layer 3 Maximum
#           # Reflectivity
#   91-92,  # Reserved for internal PUP and RPG Use
#   95,     # Composite Reflectivity  Raster Image/Nongeographic Alpha
#           # Edited for AP
#   96,     # Composite Reflectivity  Raster Image/Nongeographic Alpha
#           # Edited for AP
#   97,     # Composite Reflectivity  Raster Image/Nongeographic Alpha
#           # Edited for AP
#   98,     # Composite Reflectivity  Raster Image/Nongeographic Alpha
#           # Edited for AP
#   100,    # Site Adaptable parameters for VAD Wind Profile (Product 48)
#   101,    # Storm Track             Alphanumeric Block
#   102,    # Hail Index              Alphanumeric Block
#   103,    # Spare
#   104,    # TVS                     Alphanumeric Block
#   105,    # Site Adaptable Parameters for Combined Shear
#   106,    # Spare
#   107,    # Surface Rainfall (1 hr) Alphanumeric Block
#   108,    # Surface Rainfall (3 hr) Alphanumeric Block
#   109,    # Storm Total Rainfall    Alphanumeric Block
#           # Accumulation
#   110,    # Clutter Likelihood      Alphanumeric Block
#           # Reflectivity
#   111,    # Clutter Likelihood      Alphanumeric Block
#           # Doppler
#   112-131,# Reserved for Future Products
#   136,    # SuperOb Adaptable       Latitude, Longitude
#           #                         (ICD packet code 27)
#   139,    # Spare
#   140,    # Gust Front MIGFA        Generic Data Format
#   141,    # Mesocyclone Detection   Geographic and Non-geographic Alpha
#   143,    # Tornado Vortex          Geographic and Non-geographic Alpha
#           # Signature Rapid Update
#   149,    # Digital Mesocyclone     Generic Data Format
#           # Detection
#   152,    # Archive III Status      Product Generic Data Format
#   156,    # Eddy Dissipation Rate   Digital Radial Data Array
#   157,    # Eddy Dissipation Rate   Digital Radial Data Array
#           # Confidence
#   166,    # Melting Layer           Linked Contour Vectors/
#           #                         Set Color Level
#   176,    # Digital Instantaneous   Generic Radial Product Format
#           # Precipitation Rate
#   178-193,# Reserved for Future Products
#   196-198,# Reserved for Future Products
#   200-210,# Reserved for Future Products
#   211-220,# Reserved for Future Products
#   221-230,# Reserved for Future Products
#   231-240,# Reserved for Future Products
#   241-250,# Reserved for Future Products
#   251-260,# Reserved for Future Products
#   261-270,# Reserved for Future Products
#   271-280,# Reserved for Future Products
#   281-290,# Reserved for Future Products
#   291-296,# Reserved for Internal RPG Use.
#   297-299,# Reserved for Internal RPG Use.



"""
Radar retrievals.

"""

from .kdp_proc import kdp_maesaka, kdp_schneebeli, kdp_vulpiani
from .echo_class import steiner_conv_strat, hydroclass_semisupervised
from .echo_class import get_freq_band
from .gate_id import map_profile_to_gates, fetch_radar_time_profile
from .simple_moment_calculations import calculate_snr_from_reflectivity
from .simple_moment_calculations import calculate_velocity_texture
from .simple_moment_calculations import compute_snr, compute_l, compute_cdr
from .simple_moment_calculations import compute_noisedBZ
from .spectra_calculations import spectra_moments, dealias_spectra
from .advection import grid_displacement_pc, grid_shift
from .qpe import est_rain_rate_zpoly, est_rain_rate_z, est_rain_rate_kdp
from .qpe import est_rain_rate_a, est_rain_rate_zkdp, est_rain_rate_za
from .qpe import est_rain_rate_hydro
from .vad import velocity_azimuth_display
from .qvp import quasi_vertical_profile

__all__ = [s for s in dir() if not s.startswith('_')]



"""
Colorblind friendly radar colormaps

Available colormaps, reversed versions are also provided, these
colormaps are available within matplotlib with names pyart_COLORMAP':

    * HomeyerRainbow
"""

import warnings

import matplotlib as mpl
import matplotlib.colors as colors

from .cm import _reverser, revcmap, _reverse_cmap_spec
from ._cm_colorblind import datad, yuv_rainbow_24


def _generate_cmap(name, lutsize):
    """Generates the requested cmap from it's name *name*. The lut size is
    *lutsize*."""

    spec = datad[name]
    # Generate the colormap object.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", FutureWarning)
        if isinstance(spec, dict) and 'red' in spec.keys():
            return colors.LinearSegmentedColormap(name, spec, lutsize)
        else:
            return colors.LinearSegmentedColormap.from_list(name, spec, lutsize)

cmap_d = dict()

# reverse all the colormaps.
# reversed colormaps have '_r' appended to the name.

LUTSIZE = mpl.rcParams['image.lut']

# need this list because datad is changed in loop
_cmapnames = list(datad.keys())

# Generate the reversed specifications ...

for cmapname in _cmapnames:
    spec = datad[cmapname]
    spec_reversed = _reverse_cmap_spec(spec)
    datad[cmapname + '_r'] = spec_reversed

# Precache the cmaps with ``lutsize = LUTSIZE`` ...

# Use datad.keys() to also add the reversed ones added in the section above:
for cmapname in datad.keys():
    cmap_d[cmapname] = _generate_cmap(cmapname, LUTSIZE)

locals().update(cmap_d)

# register the colormaps so that can be accessed with the names pyart_XXX
for name, cmap in cmap_d.items():
    full_name = 'pyart_' + name
    mpl.cm.register_cmap(name=full_name, cmap=cmap)



"""
Core Py-ART classes and function for interacting with weather radar data.

"""

from .radar import Radar
from .radar_spectra import RadarSpectra
from .grid import Grid
from .wind_profile import HorizontalWindProfile

from .transforms import antenna_to_cartesian
from .transforms import antenna_vectors_to_cartesian
from .transforms import cartesian_to_geographic
from .transforms import cartesian_vectors_to_geographic
from .transforms import cartesian_to_geographic_aeqd
from .transforms import geographic_to_cartesian
from .transforms import geographic_to_cartesian_aeqd

__all__ = [s for s in dir() if not s.startswith('_')]



"""
Classes for specifying what gates are included and excluded from routines.

"""

from .gatefilter import GateFilter, moment_based_gate_filter
from .gatefilter import moment_and_texture_based_gate_filter
from .gatefilter import temp_based_gate_filter, iso0_based_gate_filter

__all__ = [s for s in dir() if not s.startswith('_')]




import pyart
import argparse

if __name__ == '__main__':

    # parse the arguments
    parser = argparse.ArgumentParser(
        description='Print information on a radar file.')
    parser.add_argument('filename', type=str, help='radar file to check')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-f', '--full', dest='level', action='store_const',
                       const='f', help='print out all information')
    group.add_argument('-s', '-standard', dest='level', action='store_const',
                       const='s', help='print out standard information')
    group.add_argument('-c', '--compact', dest='level', action='store_const',
                       const='c',
                       help='print out minimal information (default)')
    parser.set_defaults(level='c')

    igroup = parser.add_argument_group(
        title='ingest method, optional',
        description=('The method of file ingest can be specified. '
                     'If no ingest is specified, the format of the file will '
                     'be used to determine the best ingest method. '
                     'Specify only one of the following:'))

    igroup.add_argument('--sigmet', action='store_true',
                        help='Sigmet/IRIS ingest')
    igroup.add_argument('--mdv', action='store_true', help='MDV ingest')
    igroup.add_argument('--cfradial', action='store_true',
                        help='CF/Radial ingest')
    igroup.add_argument('--rsl', action='store_true',
                        help='RSL ingest')
    igroup.add_argument('--nexrad_archive', action='store_true',
                        help='NEXRAD level 2 archive ingest')
    igroup.add_argument('--nexrad_cdm', action='store_true',
                        help='NEXRAD level 2 CDM ingest')

    parser.add_argument('-v', '--version', action='version',
                        version='Py-ART version %s' % (pyart.__version__))
    args = parser.parse_args()

    # read in the file
    if args.sigmet:
        radar = pyart.io.read_sigmet(args.filename)
    elif args.mdv:
        radar = pyart.io.read_mdv(args.filename)
    elif args.cfradial:
        radar = pyart.io.read_cfradial(args.filename)
    elif args.rsl:
        radar = pyart.io.read_rsl(args.filename)
    elif args.nexrad_archive:
        radar = pyart.io.read_nexrad_archive(args.filename)
    elif args.nexrad_cdm:
        radar = pyart.io.read_nexrad_cdm(args.filename)
    else:
        radar = pyart.io.read(args.filename)

    # print out information
    radar.info(args.level)



""" Unit Tests for Py-ART's core/radar_spectra.py module. """

import inspect

import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
import pytest

try:
    import xarray as xr
    _XARRAY_AVAILABLE = True
except ImportError:
    _XARRAY_AVAILABLE = False

import pyart


@pytest.mark.skipif(not _XARRAY_AVAILABLE,
                    reason="Xarray is not installed.")
def test_gate_longitude_latitude():
    radar = pyart.testing.make_empty_spectra_radar(5, 4, 2)
    radar.ds['azimuth'] = xr.DataArray(
        np.array([0, 90, 180, 270, 0]), dims='time')
    radar.ds['elevation'] = xr.DataArray(
        np.array([0, 0, 0, 0, 10]), dims='time')
    radar.ds['range'] = xr.DataArray(
        np.array([5, 15, 25, 35]), dims='range')
    assert radar.gate_longitude.data.shape == (5, 4)
    assert radar.gate_latitude.data.shape == (5, 4)
    assert_almost_equal(radar.gate_longitude.data[0, 0], -97.5, 1)
    assert_almost_equal(radar.gate_latitude.data[0, 0], 36.5, 1)
    # reset and try again with a non-default lat_0/lon_0
    radar.ds.projection.pop('_include_lon_0_lat_0')
    radar.ds.projection['lat_0'] = 20.0
    radar.ds.projection['lon_0'] = 60.0
    radar.init_gate_longitude_latitude()
    assert_almost_equal(radar.gate_longitude.data[0, 0], 60.0, 1)
    assert_almost_equal(radar.gate_latitude.data[0, 0], 20.0, 1)


@pytest.mark.skipif(not _XARRAY_AVAILABLE,
                    reason="Xarray is not installed.")
def test_gate_altitude():
    radar = pyart.testing.make_empty_spectra_radar(5, 4, 2)
    radar.ds['azimuth'] = xr.DataArray(
        np.array([0, 90, 180, 270, 0]), dims='time')
    radar.ds['elevation'] = xr.DataArray(
        np.array([0, 0, 0, 0, 10]), dims='time')
    radar.ds['range'] = xr.DataArray(
        np.array([5, 15, 25, 35]), dims='range')

    assert radar.gate_altitude.values.shape == (5, 4)
    assert_almost_equal(radar.gate_altitude.values[0, 0], 200.0, 1)

    radar.altitude.values[0] = 150.
    radar.init_gate_altitude()
    assert_almost_equal(radar.gate_altitude.values[0, 0], 150.0, 1)


@pytest.mark.skipif(not _XARRAY_AVAILABLE,
                    reason="Xarray is not installed.")
def test_gate_x_y_z():
    radar = pyart.testing.make_empty_spectra_radar(5, 5, 2)
    radar.ds['azimuth'] = xr.DataArray(
        np.array([0, 90, 180, 270, 0]), dims='time')
    radar.ds['elevation'] = xr.DataArray(
        np.array([0, 0, 0, 0, 10]), dims='time')
    radar.ds['range'] = xr.DataArray(
        np.array([5, 15, 25, 35, 45]), dims='range')
    radar.init_gate_x_y_z()
    assert radar.gate_x.values.shape == (5, 5)
    assert_allclose(radar.gate_x.values[0], [0, 0, 0, 0, 0], atol=1e-14)
    assert_allclose(radar.gate_x.values[1], [5, 15, 25, 35, 45], atol=1e-14)
    assert_allclose(radar.gate_x.values[2], [0, 0, 0, 0, 0], atol=1e-5)
    assert_allclose(
        radar.gate_x.values[3], [-5, -15, -25, -35, -45], atol=1e-14)

    assert radar.gate_y.values.shape == (5, 5)
    assert_allclose(radar.gate_y.values[0], [5, 15, 25, 35, 45], atol=1e-14)
    assert_allclose(radar.gate_y.values[1], [0, 0, 0, 0, 0], atol=1e-5)
    assert_allclose(
        radar.gate_y.values[2], [-5, -15, -25, -35, -45], atol=1e-14)
    assert_allclose(radar.gate_y.values[3], [0, 0, 0, 0, 0], atol=1e-6)

    assert radar.gate_z.values.shape == (5, 5)
    z_sweep0 = np.array([1.47e-6, 1.324e-5, 3.679e-5, 7.210e-5, 1.1919e-4])
    assert_allclose(radar.gate_z.values[0], z_sweep0, atol=1e-3)
    assert_allclose(radar.gate_z.values[1], z_sweep0, atol=1e-3)
    assert_allclose(radar.gate_z.values[2], z_sweep0, atol=1e-3)
    assert_allclose(radar.gate_z.values[3], z_sweep0, atol=1e-3)


@pytest.mark.skipif(not _XARRAY_AVAILABLE,
                    reason="Xarray is not installed.")
def test_init_gate_x_y_z():
    radar = pyart.testing.make_empty_spectra_radar(4, 5, 1)
    radar.ds['azimuth'] = xr.DataArray(np.array([0, 90, 180, 270]), dims='time')
    radar.ds['elevation'] = xr.DataArray(np.array([0, 0, 0, 0]), dims='time')
    radar.ds['range'] = xr.DataArray(
        np.array([5, 15, 25, 35, 45]), dims='range')
    # access and check initial gate locations
    assert_allclose(
        radar.gate_x.values[1], [0., 4.3627267, 8.72545, 13.08817, 17.450884],
        atol=1e-14)
    assert_allclose(
        radar.gate_y.values[0], [0., 249.97852, 499.95685, 749.935, 999.9128],
        atol=1e-14)
    z_sweep0 = np.array([0., 3., 6., 9., 13.])
    assert_allclose(radar.gate_z.values[0], z_sweep0, atol=1e-3)

    # change range, gate_x, y, z are not updated
    radar.ds['range'] = xr.DataArray(
        np.array([15, 25, 35, 45, 55]), dims='range')
    assert_allclose(
        radar.gate_x.values[1], [0., 4.3627267, 8.72545, 13.08817, 17.450884],
        atol=1e-14)
    assert_allclose(
        radar.gate_y.values[0], [0., 249.97852, 499.95685, 749.935, 999.9128],
        atol=1e-14)
    z_sweep0 = np.array([0., 3., 6., 9., 13.])
    assert_allclose(radar.gate_z.values[0], z_sweep0, atol=1e-3)

    # call init_gate_x_y_z, now the attributes are updated
    radar.init_gate_x_y_z()
    assert_allclose(radar.gate_x.values[1], [15, 25, 35, 45, 55], atol=1e-14)
    assert_allclose(radar.gate_y.values[0], [15, 25, 35, 45, 55], atol=1e-14)
    z_sweep0 = np.array([1.324e-5, 3.679e-5, 7.210e-5, 1.1919e-4, 1.7805e-4])
    assert_allclose(radar.gate_z.values[0], z_sweep0, atol=1e-3)


@pytest.mark.skipif(not _XARRAY_AVAILABLE,
                    reason="Xarray is not installed.")
def test_rays_per_sweep_attribute():
    radar = pyart.testing.make_target_spectra_radar()
    rays_per_sweep = radar.rays_per_sweep
    assert isinstance(rays_per_sweep, xr.DataArray)
    assert rays_per_sweep.values.shape == (1, )
    assert rays_per_sweep.values[0] == 10


@pytest.mark.skipif(not _XARRAY_AVAILABLE,
                    reason="Xarray is not installed.")
def test_radar_creation():
    radar = pyart.testing.make_target_spectra_radar()
    assert isinstance(radar, pyart.core.RadarSpectra)


@pytest.mark.skipif(not _XARRAY_AVAILABLE,
                    reason="Xarray is not installed.")
def test_iterators():
    radar = pyart.testing.make_empty_spectra_radar(30, 20, 5)

    starts = [0]
    ends = [29]
    starts_ends = [(s, e) for s, e in zip(starts, ends)]

    assert inspect.isgenerator(radar.iter_start())
    assert [s for s in radar.iter_start()] == starts

    assert inspect.isgenerator(radar.iter_end())
    assert [s for s in radar.iter_end()] == ends

    assert inspect.isgenerator(radar.iter_start_end())
    assert [s for s in radar.iter_start_end()] == starts_ends

    assert inspect.isgenerator(radar.iter_slice())
    for s, start, end in zip(radar.iter_slice(), starts, ends):
        assert s.start == start
        assert s.stop == end + 1
        assert s.step is None

    assert inspect.isgenerator(radar.iter_field('spectra'))
    for d in radar.iter_field('spectra'):
        assert d.shape == (30, 20, 5)
        assert d.dtype == np.float64
    pytest.raises(KeyError, radar.iter_field, 'foobar')

    assert inspect.isgenerator(radar.iter_azimuth())
    for d in radar.iter_azimuth():
        assert d.shape == (30, )

    assert inspect.isgenerator(radar.iter_elevation())
    for d in radar.iter_elevation():
        assert d.shape == (30, )


@pytest.mark.skipif(not _XARRAY_AVAILABLE,
                    reason="Xarray is not installed.")
def test_vpt():
    radar = pyart.testing.make_target_spectra_radar()
    vpt = radar.to_vpt()
    assert 'instrument_name' in vpt.metadata
    assert vpt.fields['reflectivity']['data'].shape == (10, 20)
    assert list(vpt.fields.keys()) == [
        'reflectivity', 'velocity', 'spectrum_width',
        'skewness', 'kurtosis']
    assert vpt.range['data'].shape == (20,)
    assert vpt.time['data'].shape == (10,)



""" Unit Tests for Py-ART's io/nexrad_archive.py module using a MSG1 file. """

import warnings

import numpy as np
from numpy.testing import assert_almost_equal
from numpy.ma.core import MaskedArray
import pytest

import pyart

#######################################################
# read_nexrad_archive tests (verify radar attributes) #
#######################################################


NRAYS = 2567
NGATES = 1840


# read in the sample file, ignore warnings
with warnings.catch_warnings():
    warnings.simplefilter('ignore', category=UserWarning)
    radar = pyart.io.read_nexrad_archive(
        pyart.testing.NEXRAD_ARCHIVE_MSG1_FILE, station='KLOT')


# time attribute
def test_time():
    assert 'comment' in radar.time.keys()
    assert 'long_name' in radar.time.keys()
    assert 'standard_name' in radar.time.keys()
    assert 'units' in radar.time.keys()
    assert 'calendar' in radar.time.keys()
    assert 'data' in radar.time.keys()
    assert radar.time['units'] == 'seconds since 2003-01-01T00:09:21Z'
    assert radar.time['data'].shape == (NRAYS, )
    assert_almost_equal(radar.time['data'][1], .506, 3)


# range attribute
def test_range():
    assert 'long_name' in radar.range
    assert 'standard_name' in radar.range
    assert 'meters_to_center_of_first_gate' in radar.range
    assert 'meters_between_gates' in radar.range
    assert 'units' in radar.range
    assert 'data' in radar.range
    assert 'spacing_is_constant' in radar.range
    assert radar.range['data'].shape == (NGATES, )
    assert_almost_equal(radar.range['data'][0], -375, 0)
    assert_almost_equal(radar.range['data'][1], -125, 0)
    assert_almost_equal(radar.range['data'][-1], 459375, 0)


# metadata attribute
def test_metadata():
    assert 'instrument_name' in radar.metadata
    assert 'source' in radar.metadata
    assert 'vcp_pattern' not in radar.metadata


# scan_type attribute
def test_scan_type():
    assert radar.scan_type == 'ppi'


# latitude attribute
def test_latitude():
    assert 'data' in radar.latitude
    assert 'standard_name' in radar.latitude
    assert 'units' in radar.latitude
    assert radar.latitude['data'].shape == (1, )
    assert_almost_equal(radar.latitude['data'], 41.6, 1)


# longitude attribute
def test_longitude():
    assert 'data' in radar.longitude
    assert 'standard_name' in radar.longitude
    assert 'units' in radar.longitude
    assert radar.longitude['data'].shape == (1, )
    assert_almost_equal(radar.longitude['data'], -88.1, 1)


# altitude attribute
def test_altitude():
    assert 'data' in radar.altitude
    assert 'standard_name' in radar.altitude
    assert 'units' in radar.altitude
    assert 'positive' in radar.altitude
    assert radar.altitude['data'].shape == (1, )
    assert_almost_equal(radar.altitude['data'], 663, 0)


# altitude_agl attribute
def test_altitude_agl():
    assert radar.altitude_agl is None


# sweep_number attribute
def test_sweep_number():
    assert 'standard_name' in radar.sweep_number
    assert np.all(radar.sweep_number['data'] == range(7))


# sweep_mode attribute
def test_sweep_mode():
    assert 'standard_name' in radar.sweep_mode
    assert radar.sweep_mode['data'].shape == (7, )
    assert np.all(radar.sweep_mode['data'] == [b'azimuth_surveillance'])


# fixed_angle attribute
def test_fixed_angle():
    assert 'standard_name' in radar.fixed_angle
    assert 'units' in radar.fixed_angle
    assert radar.fixed_angle['data'].shape == (7, )
    assert_almost_equal(radar.fixed_angle['data'][0], 0.5, 1)


# sweep_start_ray_index attribute
def test_sweep_start_ray_index():
    assert 'long_name' in radar.sweep_start_ray_index
    assert radar.sweep_start_ray_index['data'].shape == (7, )
    assert_almost_equal(radar.sweep_start_ray_index['data'][1], 367, 0)


# sweep_end_ray_index attribute
def test_sweep_end_ray_index():
    assert 'long_name' in radar.sweep_end_ray_index
    assert radar.sweep_end_ray_index['data'].shape == (7, )
    assert_almost_equal(radar.sweep_end_ray_index['data'][1], 733, 0)


# target_scan_rate attribute
def test_target_scan_rate():
    assert radar.target_scan_rate is None


# azimuth attribute
def test_azimuth():
    assert 'standard_name' in radar.azimuth
    assert 'long_name' in radar.azimuth
    assert 'units' in radar.azimuth
    assert 'axis' in radar.azimuth
    assert_almost_equal(radar.azimuth['data'][0], 245.874, 3)
    assert_almost_equal(radar.azimuth['data'][10], 255.71777, 3)


# elevation attribute
def test_elevation():
    assert 'standard_name' in radar.elevation
    assert 'long_name' in radar.azimuth
    assert 'units' in radar.elevation
    assert 'axis' in radar.elevation
    assert radar.elevation['data'].shape == (NRAYS, )
    assert_almost_equal(radar.elevation['data'][0], 0.48, 2)


# scan_rate attribute
def test_scan_rate():
    assert radar.scan_rate is None


# antenna_transition attribute
def test_antenna_transition():
    assert radar.antenna_transition is None


# instrument_parameters attribute
def test_instument_parameters():
    assert 'nyquist_velocity' in radar.instrument_parameters
    nyq = radar.instrument_parameters['nyquist_velocity']['data']
    assert_almost_equal(nyq[0], 0, 0)
    assert_almost_equal(nyq[400], 28.34, 2)

    assert 'unambiguous_range' in radar.instrument_parameters
    unamb = radar.instrument_parameters['unambiguous_range']['data']
    assert_almost_equal(unamb[0], 466000., 0)
    assert_almost_equal(unamb[400], 137000., 0)


# radar_calibration attribute
def test_radar_calibration():
    assert radar.radar_calibration is None


# ngates attribute
def test_ngates():
    assert radar.ngates == NGATES


# nrays attribute
def test_nrays():
    assert radar.nrays == NRAYS


# nsweeps attribute
def test_nsweeps():
    assert radar.nsweeps == 7


####################
# fields attribute #
####################


FIELDS = ['reflectivity', 'spectrum_width', 'velocity']


@pytest.mark.parametrize("field", FIELDS)
def test_field_dics(field):
    description = "field : %s, dictionary" % field
    check_field_dic.description = description
    check_field_dic(field)


def check_field_dic(field):
    """ Check that the required keys are present in a field dictionary. """
    assert 'standard_name' in radar.fields[field]
    assert 'units' in radar.fields[field]
    assert '_FillValue' in radar.fields[field]
    assert 'coordinates' in radar.fields[field]


@pytest.mark.parametrize("field", FIELDS)
def test_field_shapes(field):
    description = "field : %s, shape" % field
    check_field_shape.description = description
    check_field_shape(field)


def check_field_shape(field):
    assert radar.fields[field]['data'].shape == (NRAYS, NGATES)


fields = {'spectrum_width': MaskedArray,
          'reflectivity': MaskedArray,
          'velocity': MaskedArray}
@pytest.mark.parametrize(
    "field, field_type", fields.items(), ids=list(fields.keys()))
def test_field_types(field, field_type):
    description = "field : %s, type" % field
    check_field_type.description = description
    check_field_type(field, field_type)


def check_field_type(field, field_type):
    assert type(radar.fields[field]['data']) is field_type


def test_field_data_spectrum_width():
    assert_almost_equal(radar.fields['spectrum_width']['data'][367, 16:19],
                        [0.0, 15.5, 16.5])


def test_field_data_velocity():
    assert_almost_equal(radar.fields['velocity']['data'][367, 16:19],
                        [-17.5, -13.0, 7.5])


def test_field_data_reflectivity():
    assert_almost_equal(
        radar.fields['reflectivity']['data'][0, 8:16],
        [1.0, 1.0, 0.4375, -0.6875, -1.8125, -2.9375, -3.5, -3.5])


def test_field_data_nearest_neighbor():
    # read in the sample file, ignore warnings
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=UserWarning)
        radar = pyart.io.read_nexrad_archive(
            pyart.testing.NEXRAD_ARCHIVE_MSG1_FILE,
            station='KLOT', linear_interp=False)
    assert_almost_equal(
        radar.fields['reflectivity']['data'][0, 8:16],
        [1.0, 1.0, 1.0, 1.0, -3.5, -3.5, -3.5, -3.5])


def test_single_scan():
    radar = pyart.io.read_nexrad_archive(
        pyart.testing.NEXRAD_ARCHIVE_MSG1_FILE,
        station='KLOT', scans=[0])
    assert radar.nrays == 367
    assert radar.ngates == 460
    assert radar.nsweeps == 1
    assert_almost_equal(radar.fields['reflectivity']['data'][0, 6:10],
                        [19, 26.5, 14.5, 21.5])
    assert_almost_equal(radar.range['data'][0], 0, 0)
    assert_almost_equal(radar.range['data'][1], 1000, 0)


def test_only_ref():
    radar = pyart.io.read_nexrad_archive(
        pyart.testing.NEXRAD_ARCHIVE_MSG1_FILE,
        station='KLOT', scans=[0, 2, 4, 5, 6],
        exclude_fields=['velocity', 'spectrum_width'])
    assert_almost_equal(radar.range['data'][0], 0, 0)
    assert_almost_equal(radar.range['data'][1], 1000, 0)
    assert_almost_equal(radar.fields['reflectivity']['data'][0, 6:10],
                        [19, 26.5, 14.5, 21.5])
    assert radar.nrays == 367 + 368 + 366 + 366 + 366
    assert radar.ngates == 460
    assert radar.nsweeps == 5



"""
Function for creating simulated velocity fields.

"""

import numpy as np
from scipy.interpolate import interp1d

from ..config import get_metadata, get_field_name


def simulated_vel_from_profile(
        radar, profile, interp_kind='linear', sim_vel_field=None):
    """
    Create simulated radial velocities from a profile of horizontal winds.

    Parameters
    ----------
    radar : Radar
        Radar instance which provides the scanning parameters for the
        simulated radial velocities.
    profile : HorizontalWindProfile
        Profile of horizontal winds.
    interp_kind : str, optional
        Specifies the kind of interpolation used to determine the winds at a
        given height. Must be one of 'linear', 'nearest', 'zero', 'slinear',
        'quadratic', or 'cubic'. The the documentation for the SciPy
        scipy.interpolate.interp1d function for descriptions.
    sim_vel_field : str, optional
        Name to use for the simulated velocity field metadata. None will use
        the default field name from the Py-ART configuration file.

    Returns
    -------
    sim_vel : dict
        Dictionary containing a radar field of simulated radial velocities.

    """
    # parse parameters
    if sim_vel_field is None:
        sim_vel_field = get_field_name('simulated_velocity')

    # radar parameters
    azimuths = np.deg2rad(radar.azimuth['data']).reshape(-1, 1)
    elevations = np.deg2rad(radar.elevation['data']).reshape(-1, 1)
    gate_altitudes = radar.gate_altitude['data']

    if isinstance(gate_altitudes, np.ma.MaskedArray):
        gate_altitudes = gate_altitudes.filled(np.nan)

    # prepare wind profile for interpolation
    if isinstance(profile.height, np.ma.MaskedArray):
        height = profile.height.filled(np.nan)
    else:
        height = profile.height

    height_is_not_nan = ~np.isnan(height)
    winds = np.empty((2, len(height)), dtype=np.float64)
    if isinstance(profile.u_wind, np.ma.MaskedArray):
        winds[0] = profile.u_wind.filled(np.nan)
    else:
        winds[0] = profile.u_wind

    if isinstance(profile.v_wind, np.ma.MaskedArray):
        winds[1] = profile.v_wind.filled(np.nan)
    else:
        winds[1] = profile.v_wind

    wind_is_not_nan = np.logical_and(~np.isnan(winds[0]), ~np.isnan(winds[1]))
    no_nans = np.logical_and(height_is_not_nan, wind_is_not_nan)
    height = height[no_nans]
    winds[0] = winds[0][no_nans]
    winds[1] = winds[1][no_nans]
    wind_interp = interp1d(
        height, winds, kind=interp_kind, bounds_error=False)

    # interpolated wind speeds at all gates altitudes
    gate_winds = wind_interp(gate_altitudes)
    gate_u = np.ma.masked_invalid(gate_winds[0])
    gate_v = np.ma.masked_invalid(gate_winds[1])

    # calculate the radial velocity for all gates
    radial_vel = (gate_u * np.sin(azimuths) * np.cos(elevations) +
                  gate_v * np.cos(azimuths) * np.cos(elevations))

    sim_vel = get_metadata(sim_vel_field)
    sim_vel['data'] = radial_vel
    return sim_vel



""" Unit Tests for Py-ART's retrieve/gate_id.py module. """

import numpy as np
import netCDF4

import pyart


def test_map_profile_to_gates():
    test_radar = pyart.testing.make_empty_ppi_radar(100, 360, 5)
    foo_field = {'data': np.zeros([360 * 5, 100])}
    test_radar.add_field('foo', foo_field)
    temp_dict = pyart.retrieve.map_profile_to_gates(
        np.ones(100), np.linspace(0, 1000, 100), test_radar)[1]
    assert temp_dict['data'].mean() == 1.0


def test_fetch_radar_time_profile():
    test_radar = pyart.testing.make_empty_ppi_radar(100, 360, 5)
    test_radar.time['units'] = 'seconds since 2011-05-10T00:00:01Z'
    test_radar.time['data'][0] = 41220.     # 2nd time in interpolated sonde

    sonde_dset = netCDF4.Dataset(pyart.testing.INTERP_SOUNDE_FILE)

    dic = pyart.retrieve.fetch_radar_time_profile(sonde_dset, test_radar)
    assert 'wdir' in dic
    assert 'wspd' in dic
    assert 'height' in dic
    assert round(dic['wdir'][0]) == 185



"""
Miscellaneous utility functions.

The location and names of these functions within Py-ART may change between
versions without depreciation, use with caution.

"""

from .circular_stats import angular_mean, angular_std
from .circular_stats import angular_mean_deg, angular_std_deg
from .circular_stats import interval_mean, interval_std
from .circular_stats import mean_of_two_angles, mean_of_two_angles_deg
from .datetime_utils import datetime_from_radar, datetimes_from_radar
from .datetime_utils import datetime_from_dataset, datetimes_from_dataset
from .datetime_utils import datetime_from_grid
from .xsect import cross_section_ppi, cross_section_rhi
from .hildebrand_sekhon import estimate_noise_hs74
from .radar_utils import is_vpt, to_vpt, join_radar
from .simulated_vel import simulated_vel_from_profile
from .sigmath import texture_along_ray, rolling_window
from .sigmath import texture, angular_texture_2d

__all__ = [s for s in dir() if not s.startswith('_')]



""" Module containing calculations for spectra moments. """

import numpy as np

from ..config import get_metadata
from ..util.hildebrand_sekhon import estimate_noise_hs74


def spectra_moments(radar):
    """ Retrieves the radar moments using a spectra radar object.

    Parameter
    ---------
    radar : RadarSpectra
        Radar spectra object to use for the calculations.

    Returns
    -------
    fields : dict
        Field dictionaries containing moment data.

    """
    field_list = {}
    times = len(radar.time.values)
    rng = len(radar.range.values)
    ref = np.zeros((times, rng))
    vel = np.zeros((times, rng))
    spec_width = np.zeros((times, rng))
    skew = np.zeros((times, rng))
    kurt = np.zeros((times, rng))
    spectra = radar.ds.spectra.values
    velocity_bins = radar.velocity_bins.values
    d_spectra = np.zeros((times, rng, len(velocity_bins)*3))
    print(spectra.shape)
    wavelength = radar.ds.attrs['wavelength']
    for i in range(times):
        if i % 100 == 0:
            print('Dealiasing %d/%d' % (i, times))
        spectra_idx = spectra[i, :, :]
        # Get the raw spectra, bins, and wavelength
        # Subtract the noise floor from the spectra. Also, return the integration limits
        noise_floor, left_limit, right_limit, the_spectra = _get_noise_floor_and_limits(
            spectra_idx, avg_window=5)
        d_spec, dealiased_bins = dealias_spectra(
            the_spectra, velocity_bins, wavelength, left_limit, right_limit)
        noise_floor = np.tile(noise_floor, (d_spec.shape[1], 1)).T
        d_spectra[i, :, :] = np.where(d_spec < noise_floor, np.nan, d_spec)
        del the_spectra, d_spec


    ref_dict = get_metadata('reflectivity')
    ref = _get_reflectivity(d_spectra, dealiased_bins, wavelength)
    ref_dict['data'] = ref
    field_list['reflectivity'] = ref_dict

    vel_dict = get_metadata('velocity')
    vel = _get_mean_velocity(d_spectra, dealiased_bins, wavelength, ref)
    vel_dict['data'] = vel
    field_list['velocity'] = vel_dict

    spec_dict = get_metadata('spectrum_width')
    spec_width = _get_spectral_width(
        d_spectra, dealiased_bins, wavelength, ref, vel)
    spec_dict['data'] = spec_width
    field_list['spectrum_width'] = spec_dict

    skew_dict = {'long_name': 'skewness',
                 'standard_name': 'skewness'}
    skew = _get_skewness(
        d_spectra, dealiased_bins, wavelength, ref, vel, spec_width)
    skew_dict['data'] = skew
    skew_dict['coordinates'] = 'elevation azimuth range'
    field_list['skewness'] = skew_dict

    kurt_dict = {'long_name': 'kurtosis',
                 'standard_name': 'kurtosis'}
    kurt = _get_kurtosis(
        d_spectra, dealiased_bins, wavelength, ref, vel, spec_width)
    kurt_dict['data'] = kurt
    kurt_dict['coordinates'] = 'elevation azimuth range'
    field_list['kurtosis'] = kurt_dict

    return field_list


def dealias_spectra(the_spectra, vel_bins, wavelength, left_limit, right_limit):
    """ Dealias a spectra.

    Parameters
    ----------
    the_spectra : array
        Spectra field data to dealias.
    vel_bins : array
        Velocity bin data.
    wavelength : float
        Spectra radar wavelength.

    Returns
    -------
    new_spectra : array
        Dealiased spectra array.
    new_bins : array
        New velocity bins from dealiased spectra.

    """
    # Calculate mean of gate before to decide whether to dealias left
    # or right side of spectrum - continuity check!
    ref = _get_reflectivity(the_spectra, vel_bins, wavelength)
    mean_vel = _get_mean_velocity(the_spectra, vel_bins, wavelength, ref)
    mean_vel = np.array(mean_vel)
    new_bins = np.concatenate(
        [vel_bins-2*vel_bins[-1], vel_bins, vel_bins+2*vel_bins[-1]])
    # Expand interval to -3Vn, 3Vn
    n_pts = len(vel_bins)
    new_spectra = np.nan*np.ones((the_spectra.shape[0], n_pts*3))
    dealiased_already = np.zeros(the_spectra.shape[0])
    # Test for aliasing - look for tail on both right interval and left interval
    for i in range(the_spectra.shape[0]):
        # First test to dealias: peaks on both sides
        # Second test: Discontinuity in velocities
        if i > 1:
            second_vel = mean_vel[i-1]
        else:
            second_vel = mean_vel[i]
        if (np.isfinite(the_spectra[i, 0]) and np.isfinite(the_spectra[i, -1])):
            noise_region = np.where(np.isnan(the_spectra[i]))[0]
            if second_vel < 0:
                right_tail_len = int(the_spectra.shape[1]-noise_region[-1])
                new_spectra[i, n_pts-right_tail_len:n_pts] = the_spectra[
                    i, n_pts-right_tail_len:n_pts]
                new_spectra[i, n_pts:2*n_pts-right_tail_len] = the_spectra[
                    i, 0:n_pts-right_tail_len]
            else:
                left_tail_len = int(noise_region[0])
                new_spectra[i, 2*n_pts:2*n_pts+left_tail_len] = the_spectra[
                    i, 0:left_tail_len]
                new_spectra[i, n_pts+left_tail_len:2*n_pts] = the_spectra[
                    i, left_tail_len:]
            dealiased_already[i] = 1

    # Do a second check for continuity
    mean_vel = _get_mean_velocity(new_spectra, new_bins, wavelength, ref)

    for i in range(new_spectra.shape[0]):
        # First test to dealias: peaks on both sides
        # Second test: Discontinuity in velocities
        if i > 1:
            second_vel = mean_vel[i-1]
        else:
            second_vel = mean_vel[i]
        # Discontinuity = more than 1 Nyquist switch in mean velocity
        if(abs(mean_vel[i]-second_vel) > vel_bins[-1] and dealiased_already[i] == 0):
            noise_region = np.where(np.isnan(the_spectra[i]))[0]
            if mean_vel[i] > 0:
                right_tail_len = int(the_spectra.shape[1]-noise_region[-1])
                new_spectra[i, n_pts-right_tail_len:n_pts] = the_spectra[
                    i, n_pts-right_tail_len:n_pts]
                new_spectra[i, n_pts:2*n_pts-right_tail_len] = the_spectra[
                    i, 0:n_pts-right_tail_len]
            else:
                left_tail_len = int(noise_region[0])
                new_spectra[i, 2*n_pts:2*n_pts+left_tail_len] = the_spectra[
                    i, 0:left_tail_len]
                new_spectra[i, n_pts+left_tail_len:2*n_pts] = the_spectra[
                    i, left_tail_len:]
            mean_vel[i] = _get_mean_velocity(
                new_spectra[i], new_bins, wavelength, ref[i])
            dealiased_already[i] = 1
        elif dealiased_already[i] == 0:
            new_spectra[i, n_pts:2*n_pts] = the_spectra[i]
    return new_spectra, new_bins


def _get_limits_dealiased_spectra(the_spectra):
    """ Calculates limits for a dealiased spectra. """
    left = np.zeros(the_spectra.shape[0])
    right = np.zeros(the_spectra.shape[0])
    new_spec = the_spectra.copy()
    for i in range(the_spectra.shape[0]):
        try:
            peak = np.nanargmax(the_spectra[i])
            j = peak
            while(j > 0 and np.isfinite(the_spectra[i, j])):
                j = j - 1
            left[i] = j
            j = peak
            while(j < the_spectra.shape[1]-1 and np.isfinite(the_spectra[i, j])):
                j = j + 1
            right[i] = j
            new_spec[i, 0:int(left[i])-1] = np.nan
            new_spec[i, int(right[i])+1:-1] = np.nan
        except ValueError:
            left[i] = np.nan
            right[i] = np.nan
    return left, right, new_spec


def _get_reflectivity(spectra, bins, wavelength):
    """ Calculates reflectivity from a RadarSpectra object. """
    spectra_linear = 10**(spectra/10)
    radar_constant = 1e18*wavelength**4/(0.93*np.pi**5)
    if len(spectra_linear.shape) == 2:
        spec_med = radar_constant*(
            spectra_linear[:, :-1]+spectra_linear[:, 1:])/2
        diffs = np.tile(np.diff(bins), (spectra.shape[0], 1))
    elif len(spectra_linear.shape) == 3:
        spec_med = radar_constant*(
            spectra_linear[:, :, :-1]+spectra_linear[:, :, 1:])/2
        diffs = np.tile(np.diff(bins), (spectra.shape[0], spectra.shape[1], 1))
    else:
        spec_med = radar_constant*(spectra_linear[:-1]+spectra_linear[1:])/2
        diffs = np.diff(bins)
    ref = np.nansum(spec_med*diffs, axis=-1)
    return 10*np.log10(ref)


def _get_mean_velocity(spectra, bins, wavelength, ref):
    """ Calculates mean velocity from a RadarSpectra object. """
    spectra_linear = 10**(spectra/10)
    radar_constant = 1e18*wavelength**4/(0.93*np.pi**5)
    ref = 10**(ref/10)
    if len(spectra_linear.shape) == 2:
        spec_med = radar_constant*(
            spectra_linear[:, :-1]+spectra_linear[:, 1:])/2
        diffs = np.tile(np.diff(bins), (spectra.shape[0], 1))
    elif len(spectra_linear.shape) == 3:
        spec_med = radar_constant*(
            spectra_linear[:, :, :-1]+spectra_linear[:, :, 1:])/2
        diffs = np.tile(np.diff(bins), (spectra.shape[0], spectra.shape[1], 1))
    else:
        spec_med = radar_constant*(spectra_linear[:-1]+spectra_linear[1:])/2
        diffs = np.diff(bins)
    bins_med = (bins[:-1]+bins[1:])/2.0
    if len(spectra_linear.shape) == 2:
        bins_med = np.tile(bins_med, (spectra.shape[0], 1))
    elif len(spectra_linear.shape) == 3:
        bins_med = np.tile(bins_med, (spectra.shape[0], spectra.shape[1], 1))
    mean_vel = np.nansum(spec_med * bins_med * diffs, axis=-1)/ref
    return mean_vel


def _get_spectral_width(spectra, bins, wavelength, ref, mean_vel):
    """ Calculates reflectivity from a RadarSpectra object. """
    spectra_linear = 10**(spectra/10)
    radar_constant = 1e18*wavelength**4/(0.93*np.pi**5)
    ref = 10**(ref/10)
    if len(spectra_linear.shape) == 2:
        spec_med = radar_constant*(
            spectra_linear[:, :-1]+spectra_linear[:, 1:])/2
        diffs = np.tile(np.diff(bins), (spectra.shape[0], 1))
        mean_vel = np.tile(mean_vel, (len(bins)-1, 1)).T
    elif len(spectra_linear.shape) == 3:
        spec_med = radar_constant*(
            spectra_linear[:, :, :-1]+spectra_linear[:, :, 1:])/2
        diffs = np.tile(np.diff(bins), (spectra.shape[0], spectra.shape[1], 1))
        mean_vel = np.tile(mean_vel.T, (len(bins)-1, 1, 1)).T
    else:
        spec_med = radar_constant*(spectra_linear[:-1]+spectra_linear[1:])/2
        diffs = np.diff(bins)
        diffs = np.tile(np.diff(bins), (spectra.shape[0], spectra.shape[1], 1))
    bins_med = (bins[:-1]+bins[1:])/2.0
    if len(spectra_linear.shape) == 2:
        bins_med = np.tile(bins_med, (spectra.shape[0], 1))
    elif len(spectra_linear.shape) == 3:
        bins_med = np.tile(bins_med, (spectra.shape[0], spectra.shape[1], 1))
    spec_wid = np.nansum(spec_med*(bins_med - mean_vel)**2*diffs, axis=-1)/ref
    return np.sqrt(spec_wid)


def _get_skewness(spectra, bins, wavelength, ref, mean_vel, spec_width):
    """ Calculates skewness from a RadarSpectra object. """
    spectra_linear = 10**(spectra/10)
    radar_constant = 1e18*wavelength**4/(0.93*np.pi**5)
    ref = 10**(ref/10)
    if len(spectra_linear.shape) == 2:
        spec_med = radar_constant*(
            spectra_linear[:, :-1]+spectra_linear[:, 1:])/2
        diffs = np.tile(np.diff(bins), (spectra.shape[0], 1))
        mean_vel = np.tile(mean_vel, (len(bins)-1, 1)).T
    elif len(spectra_linear.shape) == 3:
        spec_med = radar_constant*(
            spectra_linear[:, :, :-1]+spectra_linear[:, :, 1:])/2
        diffs = np.tile(np.diff(bins), (spectra.shape[0], spectra.shape[1], 1))
        mean_vel = np.tile(mean_vel.T, (len(bins)-1, 1, 1)).T
    else:
        spec_med = radar_constant*(spectra_linear[:-1]+spectra_linear[1:])/2
        diffs = np.diff(bins)
        diffs = np.tile(np.diff(bins), (spectra.shape[0], spectra.shape[1], 1))
    bins_med = (bins[:-1]+bins[1:])/2.0
    if len(spectra_linear.shape) == 2:
        bins_med = np.tile(bins_med, (spectra.shape[0], 1))
    elif len(spectra_linear.shape) == 3:
        bins_med = np.tile(bins_med, (spectra.shape[0], spectra.shape[1], 1))
    skew = np.nansum(spec_med*(bins_med-mean_vel)**3*diffs, axis=-1)/ref
    return skew/spec_width**3


def _get_kurtosis(spectra, bins, wavelength, ref, mean_vel, spec_width):
    """ Calculates a Kurtosis field from a RadarSpectra object. """
    spectra_linear = 10**(spectra/10)
    radar_constant = 1e18*wavelength**4/(0.93*np.pi**5)
    ref = 10**(ref/10)
    if len(spectra_linear.shape) == 2:
        spec_med = radar_constant*(
            spectra_linear[:, :-1]+spectra_linear[:, 1:])/2
        diffs = np.tile(np.diff(bins), (spectra.shape[0], 1))
        mean_vel = np.tile(mean_vel, (len(bins)-1, 1)).T
    elif len(spectra_linear.shape) == 3:
        spec_med = radar_constant*(
            spectra_linear[:, :, :-1]+spectra_linear[:, :, 1:])/2
        diffs = np.tile(np.diff(bins), (spectra.shape[0], spectra.shape[1], 1))
        mean_vel = np.tile(mean_vel.T, (len(bins)-1, 1, 1)).T
    else:
        spec_med = radar_constant*(spectra_linear[:-1]+spectra_linear[1:])/2
        diffs = np.diff(bins)
    bins_med = (bins[:-1]+bins[1:])/2.0
    if len(spectra_linear.shape) == 2:
        bins_med = np.tile(bins_med, (spectra.shape[0], 1))
    elif len(spectra_linear.shape) == 3:
        bins_med = np.tile(bins_med, (spectra.shape[0], spectra.shape[1], 1))
    kurt = np.nansum(spec_med*(bins_med - mean_vel)**4*diffs, axis=-1)/ref
    return kurt/spec_width**4


def _get_noise_floor_and_limits(the_spectra, avg_window=1):
    """ Calculates noise floor and limits from a RadarSpectra object. """
    lin_spectra = 10**(the_spectra/10.0)
    if avg_window > 1:
        for i in range(lin_spectra.shape[0]):
            lin_spectra[i] = np.convolve(
                lin_spectra[i], np.ones((avg_window))/avg_window,
                mode='same')
    noise_floor_thresh = np.zeros((lin_spectra.shape[0],))

    for i in range(lin_spectra.shape[0]):
        noise_floor = estimate_noise_hs74(
            lin_spectra[i], navg=avg_window)
        noise_floor_thresh[i] = 10*np.log10(noise_floor[0])
        the_spectra[i] = 10*np.log10(lin_spectra[i] - noise_floor[0])
    the_spectra[lin_spectra < 0] = np.nan
    left = np.nan*np.ones(the_spectra.shape[0])
    right = np.nan*np.ones(the_spectra.shape[0])
    for i in range(the_spectra.shape[0]):
        try:
            peak_loc = np.nanargmax(the_spectra[i])
            j = peak_loc
            while(np.isfinite(the_spectra[i, j]) and j > 0):
                j = j - 1
            left[i] = j
            j = peak_loc
            while(
                    np.isfinite(the_spectra[i, j]) and j < the_spectra.shape[1]-1):
                j = j + 1
            right[i] = j
        except ValueError:
            left[i] = np.nan
            right[i] = np.nan
    spectra = np.ma.masked_where(np.isnan(the_spectra), the_spectra)
    return noise_floor_thresh, left, right, spectra



def configuration(parent_package='', top_path=None):
    global config
    from numpy.distutils.misc_util import Configuration
    config = Configuration('retrieve', parent_package, top_path)
    config.add_data_dir('tests')

    # KDP processing Cython extension
    config.add_extension('_kdp_proc', sources=['_kdp_proc.c'])
    return config

if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())



""" Unit tests for the hildebrand_sekhon module. """

from pyart.util import hildebrand_sekhon

import numpy as np
from numpy.testing import assert_almost_equal

SAMPLE_SPECTRUM = np.array(
    [11.089, 10.386, 11.081, 9.696, 9.21, 11.103, 9.425,
     10.222, 8.393, 11.67, 9.632, 10.096, 9.807, 10.374,
     8.065, 8.682, 10.164, 320.206, 480.898, 640.818, 490.483,
     318.578, 9.682, 10.035, 9.134, 8.554, 9.274, 9.463,
     9.501, 10.036, 9.211, 8.265])


def test_hildebrand_sekhon():
    noise_params = hildebrand_sekhon.estimate_noise_hs74(SAMPLE_SPECTRUM)
    mean, thresh, var, nnoise = noise_params
    assert_almost_equal(mean, 10, 0)
    assert_almost_equal(thresh, 12, 0)
    assert_almost_equal(var, 0.79, 2)
    assert nnoise == 27



"""
Configuration file for the Python ARM Radar Toolkit (Py-ART)

The values for a number of Py-ART parameters and the default metadata created
when reading files, correcting fields, etc. is controlled by this single
Python configuration file.

Py-ART's configuration can be modified by setting the environmental
variable PYART_CONFIG to point to a configuration file with formatting similar
to this file.

The recommended method for changing these defaults is for users to copy this
file into their home directory, rename it to .pyart_config.py, make any
desired changes, and adjust their login scripts to set the PYART_CONFIG
environmental variable to point to .pyart_config.py in their home directory.

Py-ART's configuration can also be modified within a script or shell session
using the load_config functions, such modification will last until a the end
of the script/session or until a new configuration is loaded.

"""

# THIS FILE WAS EDITED FOR USE IN PY-ART UNIT TESTS, changed noted by
# XXX changed

##############################################################################
##############################################################################
# Simple configuration
#
# Adjust the values of the variable (left hand side of the equal sign) in this
# section for an easy method of customizing Py-ART.  Do not change the
# variables names (the right hand side of the equal sign).  More advanced
# settings are based upon these variables.  Most users will find that
# adjusting this section is all that is needed.
##############################################################################
##############################################################################

# The default fill value for masked arrays and _FillValue keys
fill_value = -5555.0        # XXX changed

# Field names used when reading in radar files and in the various correction
# and retrieval algorithms. The comments in this section provide additional
# information about the fields in that section.

# radar reflectivity fields,
reflectivity = 'REF'        # XXX changed
corrected_reflectivity = 'corrected_reflectivity'
total_power = 'total_power'

# mean doppler velocity fields, VEL
velocity = 'velocity'
corrected_velocity = 'corrected_velocity'

# spectral width field, SW
spectrum_width = 'spectrum_width'

# differential reflectivity fields, ZDR
differential_reflectivity = 'differential_reflectivity'
corrected_differential_reflectivity = 'corrected_differential_reflectivity'

# cross correlation ratio, correlation coefficient, RhoHV
cross_correlation_ratio = 'cross_correlation_ratio'

# normalized coherent power, signal quality index, SQI, NCP
normalized_coherent_power = 'normalized_coherent_power'

# Differential phase shift, PhiDP
differential_phase = 'differential_phase'
unfolded_differential_phase = 'unfolded_differential_phase'
corrected_differential_phase = 'corrected_differential_phase'

# Specific differential phase shift, KDP
specific_differential_phase = 'specific_differential_phase'
corrected_specific_differential_phase = 'corrected_specific_differential_phase'

# Linear depolarization ration (h - horizontal, v - vertical), LDR
linear_depolarization_ratio = 'linear_polarization_ratio'
linear_depolarization_ratio_h = 'linear_polarization_ratio_h'
linear_depolarization_ratio_v = 'linear_polarization_ratio_v'

# Misc fields
signal_to_noise_ratio = 'signal_to_noise_ratio'
rain_rate = 'rain_rate'
radar_estimated_rain_rate = 'radar_estimated_rain_rate'
radar_echo_classification = 'radar_echo_classification'
specific_attenuation = 'specific_attenuation'

# End of Simple Configuration section

##############################################################################
##############################################################################
# Advanced Configuration
#
# Most users will not want to make any changes in this section.  For users
# who want a more fined grained control over Py-ART's configuration this
# section provides access to these controls.  The layout of this section can
# be changed, the only requirement for a valid configuration file is that
# the ALL CAPITALIZED variable must must be present with the formatting
# present in this file.  These required variables are:
#
# FILL_VALUE, DEFAULT_METADATA, FILE_SPECIFIC_METADATA, FIELD_MAPPINGS,
# DEFAULT_FIELD_NAMES
#
# This section makes generous use of the variables in the Simple Configuration
# section, this is not required, but simplifies and enforces uniformity on
# the configuration.
##############################################################################
##############################################################################


##############################################################################
# Parameters
#
# Various parameters used in Py-ART.
##############################################################################

FILL_VALUE = fill_value     # the default fill value for masked arrays and
                            # the _FillValue key.

# The DEFAULT_FIELD_NAMES controls the field names which are used in
# by the correction and retrieval algorithms in Py-ART.  The keys of the
# dictionary are "internal" names which cannot change, the values are the field
# names which will be used in the algorithms by default.
# which will be used.  For best results use the names defined by the variables
# in simple configuration section which are also used in the DEFAULT_METADATA
# and FIELD_MAPPINGS variable.  If you choose to change a field name
# the names should also be changed in the DEFAULT_METADATA and FIELD_MAPPINGS
# variable. This is not required but highly suggested.

DEFAULT_FIELD_NAMES = {
    # internal field name (do not change): field name used (can change)
    'reflectivity': reflectivity,
    'corrected_reflectivity': corrected_reflectivity,
    'total_power': total_power,
    'velocity': velocity,
    'corrected_velocity': corrected_velocity,
    'spectrum_width': spectrum_width,
    'differential_reflectivity': differential_reflectivity,
    'corrected_differential_reflectivity':
    corrected_differential_reflectivity,
    'cross_correlation_ratio': cross_correlation_ratio,
    'normalized_coherent_power': normalized_coherent_power,
    'differential_phase': differential_phase,
    'unfolded_differential_phase': unfolded_differential_phase,
    'corrected_differential_phase': corrected_differential_phase,
    'specific_differential_phase': specific_differential_phase,
    'corrected_specific_differential_phase':
    corrected_specific_differential_phase,
    'linear_depolarization_ratio': linear_depolarization_ratio,
    'linear_depolarization_ratio_h': linear_depolarization_ratio_h,
    'linear_depolarization_ratio_v': linear_depolarization_ratio_v,
    'signal_to_noise_ratio': signal_to_noise_ratio,
    'rain_rate': rain_rate,
    'radar_estimated_rain_rate': radar_estimated_rain_rate,
    'radar_echo_classification': radar_echo_classification,
    'specific_attenuation': specific_attenuation,
}


##############################################################################
# Default metadata
#
# The DEFAULT_METADATA dictionary contains dictionaries which provide
# the default radar attribute and field metadata.  When reading in a file
# with Py-ART the FILE_SPECIFIC_METADATA variable is first queued for a
# metadata dictionary, if it is not found then the metadata in
# DEFAULT_METADATA is utilized.
##############################################################################

DEFAULT_METADATA = {

    # metadata for radar attributes
    # these closely follow the CF/Radial standard
    'azimuth': {
        'units': 'foo',         # XXX changed
        'standard_name': 'beam_azimuth_angle',
        'long_name': 'azimuth_angle_from_true_north',
        'axis': 'radial_azimuth_coordinate',
        'comment': 'Azimuth of antenna relative to true north'},

    'elevation': {
        'units': 'degrees',
        'standard_name': 'beam_elevation_angle',
        'long_name': 'elevation_angle_from_horizontal_plane',
        'axis': 'radial_elevation_coordinate',
        'comment': 'Elevation of antenna relative to the horizontal plane'},

    'range': {
        'units': 'meters',
        'standard_name': 'projection_range_coordinate',
        'long_name': 'range_to_measurement_volume',
        'axis': 'radial_range_coordinate',
        'spacing_is_constant': 'true',
        'comment': (
            'Coordinate variable for range. Range to center of each bin.')},

    'time': {
        'units': 'seconds',
        'standard_name': 'time',
        'long_name': 'time_in_seconds_since_volume_start',
        'calendar': 'gregorian',
        'comment': ('Coordinate variable for time. '
                    'Time at the center of each ray, in fractional seconds '
                    'since the global variable time_coverage_start')},

    'sweep_mode': {
        'units': 'unitless',
        'standard_name': 'sweep_mode',
        'long_name': 'Sweep mode',
        'comment': ('Options are: "sector", "coplane", "rhi", '
                    '"vertical_pointing", "idle", "azimuth_surveillance", '
                    '"elevation_surveillance", "sunscan", "pointing", '
                    '"manual_ppi", "manual_rhi"')},

    'sweep_number': {
        'units': 'count',
        'standard_name': 'sweep_number',
        'long_name': 'Sweep number'},

    'metadata': {
        'Conventions': 'CF/Radial instrument_parameters',
        'version': '1.3',
        'title': '',
        'institution': '',
        'references': '',
        'source': '',
        'history': '',
        'comment': '',
        'instrument_name': ''},

    # metadata for radar sweep information dictionaries
    'sweep_start_ray_index': {
        'long_name': 'Index of first ray in sweep, 0-based',
        'units': 'count'},

    'sweep_end_ray_index': {
        'long_name': 'Index of last ray in sweep, 0-based',
        'units': 'count'},

    'fixed_angle': {
        'long_name': 'Target angle for sweep',
        'units': 'degrees',
        'standard_name': 'target_fixed_angle'},

    # metadata for radar location attributes
    'latitude': {
        'long_name': 'Latitude',
        'standard_name': 'Latitude',
        'units': 'degrees_north'},

    'longitude': {
        'long_name': 'Longitude',
        'standard_name': 'Longitude',
        'units': 'degrees_east'},

    'altitude': {
        'long_name': 'Altitude',
        'standard_name': 'Altitude',
        'units': 'meters',
        'positive': 'up'},

    # metadata for instrument_parameter dictionary
    'prt_mode': {
        'comments': ('Pulsing mode Options are: "fixed", "staggered", '
                     '"dual". Assumed "fixed" if missing.'),
        'meta_group': 'instrument_parameters',
        'long_name': 'Pulsing mode',
        'units': 'unitless'},

    'nyquist_velocity': {
        'units': 'meters_per_second',
        'comments': "Unambiguous velocity",
        'meta_group': 'instrument_parameters',
        'long_name': 'Nyquist velocity'},

    'prt': {
        'units': 'seconds',
        'comments': ("Pulse repetition time. For staggered prt, "
                     "also see prt_ratio."),
        'meta_group': 'instrument_parameters',
        'long_name': 'Pulse repetition time'},

    'unambiguous_range': {
        'units': 'meters',
        'comments': 'Unambiguous range',
        'meta_group': 'instrument_parameters',
        'long_name': 'Unambiguous range'},

    # reflectivity fields
    reflectivity: {
        'units': 'dBZ',
        'standard_name': 'equivalent_reflectivity_factor',
        'long_name': 'Reflectivity',
        'coordinates': 'elevation azimuth range'},

    corrected_reflectivity: {
        'units': 'dBZ',
        'standard_name': 'corrected_equivalent_reflectivity_factor',
        'long_name': 'Corrected reflectivity',
        'coordinates': 'elevation azimuth range'},

    total_power: {
        'units': 'dBZ',
        'standard_name': 'equivalent_reflectivity_factor',
        'long_name': 'Total power',
        'coordinates': 'elevation azimuth range'},

    # velocity fields
    velocity: {
        'units': 'meters_per_second',
        'standard_name': 'radial_velocity_of_scatterers_away_from_instrument',
        'long_name': 'Mean dopper velocity',
        'coordinates': 'elevation azimuth range'},

    corrected_velocity: {
        'units': 'meters_per_second',
        'standard_name': (
            'corrected_radial_velocity_of_scatterers_away_from_instrument'),
        'long_name': 'Corrected mean doppler velocity',
        'coordinates': 'elevation azimuth range'},

    # spectrum width fields
    spectrum_width: {
        'units': 'meters_per_second',
        'standard_name': 'doppler_spectrum_width',
        'long_name': 'Doppler spectrum width',
        'coordinates': 'elevation azimuth range'},

    # Dual Pol fields
    differential_reflectivity: {
        'units': 'dB',
        'standard_name': 'log_differential_reflectivity_hv',
        'long_name': 'Differential reflectivity',
        'coordinates': 'elevation azimuth range'},

    corrected_differential_reflectivity: {
        'units': 'dB',
        'standard_name': 'corrected_log_differential_reflectivity_hv',
        'long_name': 'Corrected differential reflectivity',
        'coordinates': 'elevation azimuth range'},

    cross_correlation_ratio: {
        'units': 'ratio',
        'standard_name': 'cross_correlation_ratio_hv',
        'long_name': 'Cross correlation ratio (RHOHV)',
        'valid_max': 1.0,
        'valid_min': 0.0,
        'coordinates': 'elevation azimuth range'},

    normalized_coherent_power: {
        'units': 'ratio',
        'standard_name': 'normalized_coherent_power',
        'long_name': 'Normalized coherent power',
        'valid_max': 1.0,
        'valid_min': 0.0,
        'comment': 'Also know as signal quality index (SQI)',
        'coordinates': 'elevation azimuth range'},

    differential_phase: {
        'units': 'degrees',
        'standard_name': 'differential_phase_hv',
        'long_name': 'Differential phase (PhiDP)',
        'valid_max': 180.0,
        'valid_min': -180.0,
        'coordinates': 'elevation azimuth range'},

    unfolded_differential_phase: {
        'units': 'degrees',
        'standard_name': 'differential_phase_hv',
        'long_name': 'Unfolded differential phase',
        'coordinates': 'elevation azimuth range'},

    corrected_differential_phase: {
        'units': 'degrees',
        'standard_name': 'differential_phase_hv',
        'long_name': 'Corrected differential phase',
        'coordinates': 'elevation azimuth range'},

    specific_differential_phase: {
        'units': 'degrees/km',
        'standard_name': 'specific_differential_phase_hv',
        'long_name': 'Specific differential phase (KDP)',
        'coordinates': 'elevation azimuth range'},

    corrected_specific_differential_phase: {
        'units': 'degrees/km',
        'standard_name': 'specific_differential_phase_hv',
        'long_name': 'Corrected specific differential phase (KDP)',
        'coordinates': 'elevation azimuth range'},


    # depolarization ratio fields
    linear_depolarization_ratio: {
        'units': 'dB',
        'standard_name': 'log_linear_depolarization_ratio_hv',
        'long_name': 'Linear depolarization ratio',
        'coordinates': 'elevation azimuth range'},

    linear_depolarization_ratio_h: {
        'units': 'dB',
        'standard_name': 'log_linear_depolarization_ratio_h',
        'long_name': 'Linear depolarization ratio horizontal',
        'coordinates': 'elevation azimuth range'},

    linear_depolarization_ratio_v: {
        'units': 'dB',
        'standard_name': 'log_linear_depolarization_ratio_v',
        'long_name': 'Linear depolarization ratio vertical',
        'coordinates': 'elevation azimuth range'},

    # misc. fields
    signal_to_noise_ratio: {
        'units': 'dB',
        'standard_name': 'signal_to_noise_ratio',
        'long_name': 'Signal to noise ratio',
        'coordinates': 'elevation azimuth range'},

    rain_rate: {
        'units': 'kg/m2/s',
        'standard_name': 'rain_rate',
        'long_name': 'Rain rate',
        'coordinates': 'elevation azimuth range'},

    radar_estimated_rain_rate: {
        'units': 'mm/hr',
        'standard_name': 'radar_estimated_rain_rate',
        'long_name': 'Radar estimated rain rate',
        'coordinates': 'elevation azimuth range'},

    radar_echo_classification: {
        'units': 'legend',
        'standard_name': 'radar_echo_classification',
        'long_name': 'Radar Echo classification',
        'coordinates': 'elevation azimuth range'},

    specific_attenuation: {
        'units': 'dB/km',
        'standard_name': 'specific_attenuation',
        'long_name': 'Specific attenuation',
        'valid_min': 0.0,
        'valid_max': 1.0,
        'coordinates': 'elevation azimuth range'},
}


##############################################################################
# File specific metadata
#
# These dictionaries define metadata that is to be used only when reading in
# a given type of file.  This metadata is used in place of the
# DEFAULT_METADATA when it is avialable.  The main use of these variable
# is to define field specific data, it is safe to leave some/all of these
# empty if the default metadata is acceptable.
##############################################################################

# Metadata for Sigmet/IRIS files
sigmet_metadata = {}

# Metadata for NEXRAD Level II files (Archive and CDM files)
nexrad_metadata = {
    reflectivity: {
        'units': 'dBZ',
        'standard_name': 'equivalent_reflectivity_factor',
        'long_name': 'Reflectivity',
        'valid_max': 94.5,
        'valid_min': -32.0,
        'coordinates': 'elevation azimuth range'},

    velocity: {
        'units': 'meters_per_second',
        'standard_name': 'radial_velocity_of_scatterers_away_from_instrument',
        'long_name': 'Mean doppler Velocity',
        'valid_max': 95.0,
        'valid_min': -95.0,
        'coordinates': 'elevation azimuth range'},

    spectrum_width: {
        'units': 'meters_per_second',
        'standard_name': 'doppler_spectrum_width',
        'long_name': 'Spectrum Width',
        'valid_max': 63.0,
        'valid_min': -63.5,
        'coordinates': 'elevation azimuth range'},

    differential_reflectivity: {
        'units': 'dB',
        'standard_name': 'log_differential_reflectivity_hv',
        'long_name': 'log_differential_reflectivity_hv',
        'valid_max': 7.9375,
        'valid_min': -7.8750,
        'coordinates': 'elevation azimuth range'},

    differential_phase: {
        'units': 'degrees',
        'standard_name': 'differential_phase_hv',
        'long_name': 'differential_phase_hv',
        'valid_max': 360.0,
        'valid_min': 0.0,
        'coordinates': 'elevation azimuth range'},

    cross_correlation_ratio: {
        'units': 'ratio',
        'standard_name': 'cross_correlation_ratio_hv',
        'long_name': 'Cross correlation_ratio (RHOHV)',
        'valid_max': 1.0,
        'valid_min': 0.0,
        'coordinates': 'elevation azimuth range'},
}

# Metadata for CF/Radial files
cfradial_metadata = {}

# Metadata for MDV files
mdv_metadata = {'time': {'foo': 'bar'}}     # XXX changed

# Metadata for RSL files
rsl_metadata = {}

FILE_SPECIFIC_METADATA = {      # Required
    'sigmet': sigmet_metadata,
    'nexrad_archive': nexrad_metadata,
    'nexrad_cdm': nexrad_metadata,
    'cfradial': cfradial_metadata,
    'mdv': mdv_metadata,
    'rsl': rsl_metadata
}

##############################################################################
# Field name mapping
#
# These dictionaries map file field names or data types to a radar field
# name.  These are used to populate the radar.fields dictionary during a read
# in Py-ART.  A value of None will not include that field in the radar object.
# These can be over-ridden on a per-read basis using the field_mapping
# parameter, or using setting the file_field_names parameter to True.
##############################################################################

# Sigmet/IRIS file field mapping
# Note that multiple sigmet fields map to the same radar field, if
# more than one of these fields are present the radar field will be
# overwritten with the last sigmet field.
sigmet_field_mapping = {
    # Sigmet data type :field name              # (Data_type) Description
    'XHDR': None,                               # (0) Extended Header
    'DBT': total_power,                         # (1) Total Power
    'DBZ': reflectivity,                        # (2) Reflectivity
    'VEL': velocity,                            # (3) Velocity
    'WIDTH': spectrum_width,                    # (4) Width
    'ZDR': differential_reflectivity,           # (5) Diff. reflectivity
    'DBZC': corrected_reflectivity,             # (7) Corrected reflectivity
    'DBT2': total_power,                        # (8) Total Power
    'DBZ2': reflectivity,                       # (9) Reflectivity
    'VEL2': velocity,                           # (10) Velocity
    'WIDTH2': spectrum_width,                   # (11) Width
    'ZDR2': differential_reflectivity,          # (12) Diff. reflectivity
    'RAINRATE2': radar_estimated_rain_rate,     # (13) Rainfall rate
    'KDP': specific_differential_phase,         # (14) KDP (diff. phase)
    'KDP2': specific_differential_phase,        # (15) KDP (diff. phase)
    'PHIDP': differential_phase,                # (16) PhiDP (diff. phase)
    'VELC': corrected_velocity,                 # (17) Corrected velocity
    'SQI': normalized_coherent_power,           # (18) SQI
    'RHOHV': cross_correlation_ratio,           # (19) RhoHV
    'RHOHV2': cross_correlation_ratio,          # (20) RhoHV
    'DBZC2': corrected_reflectivity,            # (21) Corrected Reflec.
    'VELC2': corrected_velocity,                # (21) Corrected Velocity
    'SQI2': normalized_coherent_power,          # (23) SQI
    'PHIDP2': differential_phase,               # (24) PhiDP (diff. phase)
    'LDRH': linear_depolarization_ratio_h,      # (25) LDR xmt H, rcv V
    'LDRH2': linear_depolarization_ratio_h,     # (26) LDR xmt H, rcv V
    'LDRV': linear_depolarization_ratio_v,      # (27) LDR xmt V, rcv H
    'LDRV2': linear_depolarization_ratio_v,     # (28) LDR xmt V, rcv H
    'HEIGHT': None,                             # (32) Height (1/10 km)
    'VIL2': None,                               # (33) Linear Liquid
    'RAW': None,                                # (34) Raw Data
    'SHEAR': None,                              # (35) Wind Shear
    'DIVERGE2': None,                           # (36) Divergence
    'FLIQUID2': None,                           # (37) Floated liquid
    'USER': None,                               # (38) User type
    'OTHER': None,                              # (39) Unspecified
    'DEFORM2': None,                            # (40) Deformation
    'VVEL2': None,                              # (41) Vertical velocity
    'HVEL2': None,                              # (42) Horizontal velocity
    'HDIR2': None,                              # (43) Horiz. wind direction
    'AXDIL2': None,                             # (44) Axis of dilation
    'TIME2': None,                              # (45) Time in seconds
    'RHOH': None,                               # (46) Rho, xmt H, rcv V
    'RHOH2': None,                              # (47) Rho, xmt H, rcv V
    'RHOV': None,                               # (48) Rho, xmt V, rcv H
    'RHOV2': None,                              # (49) Rho, xmt V, rcv H
    'PHIH': None,                               # (50) Phi, xmt H, rcv V
    'PHIH2': None,                              # (51) Phi, xmt H, rcv V
    'PHIV': None,                               # (52) Phi, xmt V, rcv H
    'PHIV2': None,                              # (53) Phi, xmt V, rcv H
    'USER2': None,                              # (54) User type
    'HCLASS': radar_echo_classification,        # (55) Hydrometeor class
    'HCLASS2': radar_echo_classification,       # (56) Hydrometeor class
    'ZDRC': corrected_differential_reflectivity,
                                                # (57) Corrected diff. refl.
    'ZDRC2': corrected_differential_reflectivity,
                                                # (58) Corrected diff. refl.
}


# NEXRAD Level II Archive files
nexrad_archive_field_mapping = {
    # NEXRAD field: radar field name
    'REF': reflectivity,
    'VEL': velocity,
    'SW': spectrum_width,
    'ZDR': differential_reflectivity,
    'PHI': differential_phase,
    'RHO': cross_correlation_ratio
}

# NEXRAD Level II CDM files
nexrad_cdm_field_mapping = {
    # CDM variable name (without _HI): radar field name
    'Reflectivity': reflectivity,
    'RadialVelocity': velocity,
    'SpectrumWidth': spectrum_width,
    'DifferentialReflectivity': differential_reflectivity,
    'DifferentialPhase': differential_phase,
    'CorrelationCoefficient': cross_correlation_ratio
}

# MDV files
mdv_field_mapping = {
    # MDV moment: radar field name
    'DBZ_F': velocity,              # XXX changed
    'VEL_F': velocity,
    'WIDTH_F': spectrum_width,
    'ZDR_F': differential_reflectivity,
    'RHOHV_F': cross_correlation_ratio,
    'NCP_F': normalized_coherent_power,
    'KDP_F': specific_differential_phase,
    'PHIDP_F': differential_phase,
    'VEL_COR': corrected_velocity,
    'PHIDP_UNF': unfolded_differential_phase,
    'KDP_SOB': corrected_specific_differential_phase,
    'DBZ_AC': corrected_reflectivity, }

# CF/Radial files
cfradial_field_mapping = {}

# RSL files
# Note that multiple RSL field map to the same radar field, if
# more than one of these fields are present in the RSL data structure
# the radar field will be overwritten with the last field.
rsl_field_mapping = {
    # RSL 2 letter field: radar field           # RSL description
    'DZ': reflectivity,                         # reflectivity
    'VR': velocity,                             # velocity
    'SW': spectrum_width,                       # spectrum width
    'CZ': corrected_reflectivity,               # corrected reflectivity
    'ZT': reflectivity,                         # uncorrected reflectivity
    'DR': differential_reflectivity,            # differential reflectivity
    'LR': differential_reflectivity,            # another diff. reflectivity
    'ZD': differential_reflectivity,            # another diff. reflectivity
    'DM': None,                                 # received power
    'RH': cross_correlation_ratio,              # RhoHV
    'PH': differential_phase,                   # PhiDP
    'XZ': None,                                 # X-band reflectivity
    'CD': corrected_differential_reflectivity,  # Corrected DR.
    'MZ': None,                                 # DZ mask
    'MD': None,                                 # DR Mask
    'ZE': corrected_reflectivity,               # edited reflectivity
    'VE': corrected_velocity,                   # edited velocity
    'KD': specific_differential_phase,          # specific diff. phase
    'TI': None,                                 # TIME (unknown)
    'DX': None,                                 # ???
    'CH': None,                                 # ???
    'AH': None,                                 # ???
    'CV': None,                                 # ???
    'AV': None,                                 # ???
    'SQ': normalized_coherent_power,            # Signal Quality Index (sigmet)
    'VS': None,                                 # Radial Vel. combined
    'VL': None,                                 # Radial Vel. combined
    'VG': None,                                 # Radial Vel. combined
    'VT': None,                                 # Radial Vel. combined
    'NP': normalized_coherent_power,            # Normalized Coherent Power
    'HC': radar_echo_classification,            # Hydroclass
    'VC': None,                                 # Radial Vel. Corrected.
    'V2': None,                                 # Radial Vel cut 2
    'S2': None,                                 # Spectrum width cut 2
    'V3': None,                                 # Radial Vel cut 3
    'S3': None,                                 # Spectrum width cut 3
}

FIELD_MAPPINGS = {                  # Required variable
    'sigmet': sigmet_field_mapping,
    'nexrad_archive': nexrad_archive_field_mapping,
    'nexrad_cdm': nexrad_cdm_field_mapping,
    'cfradial': cfradial_field_mapping,
    'mdv': mdv_field_mapping,
    'rsl': rsl_field_mapping,
}



""" Unit Tests for Py-ART's io/mdv_common.py module. """

import warnings
from datetime import datetime
from io import BytesIO

import numpy as np
from numpy.testing import assert_almost_equal
import pytest

import pyart
from pyart.io.mdv_common import MdvFile

#################
# MdvFile tests #
#################


mdvfile = MdvFile(pyart.testing.MDV_PPI_FILE)


def test_master_header():
    # test the master header
    ref_master_header = {
        'chunk_hdr_offset': 2464,
        'data_collection_type': 0,
        'data_dimension': 0,
        'data_ordering': 0,
        'data_set_info': 'MDV radar volume file created by Dsr2Vol.',
        'data_set_name': 'C-SAPR',
        'data_set_source': 'ARM SGP C-SAPR',
        'field_grids_differ': 0,
        'field_hdr_offset': 1024,
        'grid_orientation': 1,
        'index_number': 611,
        'native_vlevel_type': 9,
        'nchunks': 3,
        'nfields': 1,
        'max_nx': 110,
        'max_ny': 360,
        'max_nz': 1,
        'num_data_times': 1,
        'record_len1': 1016,
        'record_len2': 1016,
        'revision_number': 1,
        'sensor_alt': 0.32760000228881836,
        'sensor_lat': 36.79615783691406,
        'sensor_lon': -97.45054626464844,
        'struct_id': 14142,
        'time_begin': 1305889260,
        'time_centroid': 1305889595,
        'time_end': 1305889595,
        'time_expire': 1305890265,
        'time_gen': 1305889595,
        'time_written': 1305889668,
        'unused_fl3212': (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0, 0.0, 0.0),
        'unused_si325': (0, 0, 0, 0, 0),
        'user_data': 0,
        'user_data_fl326': (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
        'user_data_si328': (0, 0, 0, 0, 0, 0, 0, 0),
        'user_time': 0,
        'vlevel_hdr_offset': 1440,
        'vlevel_included': 1,
        'vlevel_type': 9}

    for k, v in ref_master_header.items():
        print("checking key:", k)
        assert mdvfile.master_header[k] == v


def test_field_header():
    # test a few of the field_header
    assert len(mdvfile.field_headers) == 1
    assert mdvfile.field_headers[0]['bad_data_value'] == 0.0
    assert mdvfile.field_headers[0]['dz_constant'] == 0
    assert mdvfile.field_headers[0]['encoding_type'] == 2
    assert mdvfile.field_headers[0]['field_name'] == "DBZ_F"
    assert mdvfile.field_headers[0]['nz'] == 1


def test_vlevel_headers():
    # test the vlevel_headers
    assert len(mdvfile.vlevel_headers) == 1
    assert mdvfile.vlevel_headers[0]['record_len1'] == 1016
    assert mdvfile.vlevel_headers[0]['record_len2'] == 1016
    assert mdvfile.vlevel_headers[0]['unused_fl32'] == (0.0, 0.0, 0.0,
                                                        0.0, 0.0)
    assert mdvfile.vlevel_headers[0]['unused_si32'] == (0, 0, 0, 0)
    assert mdvfile.vlevel_headers[0]['struct_id'] == 14144
    assert len(mdvfile.vlevel_headers[0]['level']) == 122
    assert mdvfile.vlevel_headers[0]['level'][0] == 0.75
    assert mdvfile.vlevel_headers[0]['level'][10] == 11.699999809265137
    assert mdvfile.vlevel_headers[0]['level'][17] == 0.0
    assert len(mdvfile.vlevel_headers[0]['type']) == 122
    assert mdvfile.vlevel_headers[0]['type'][0:17] == (9,) * 17
    assert mdvfile.vlevel_headers[0]['type'][18] == 0


def test_chunk_headers():
    # test chunk_headers
    assert len(mdvfile.chunk_headers) == 3
    ref_chunk_header_0 = {
        'chunk_data_offset': 68580,
        'chunk_id': 3,
        'info': 'DsRadar params',
        'record_len1': 504,
        'record_len2': 504,
        'size': 240,
        'struct_id': 14145,
        'unused_si32': (0, 0)}
    for k, v in ref_chunk_header_0.items():
        print("checking key:", k)
        assert mdvfile.chunk_headers[0][k] == v


def test_calib_info():
    # test calib_info
    ref_calib_info = {
        'I0_h_co_dbm': 0.0,
        'I0_h_cx_dbm': 0.0,
        'I0_v_co_dbm': 0.0,
        'I0_v_cx_dbm': 0.0,
        'antenna_gain_h_db': 44.5,
        'antenna_gain_v_db': 44.5,
        'beamwidth_h_deg': 1.0,
        'beamwidth_v_deg': 1.0,
        'coupler_fwd_loss_h_db': 50.0,
        'coupler_fwd_loss_v_db': 50.0,
        'day': 15,
        'filter_loss_db': 1.600000023841858,
        'hour': 18,
        'ldr_h_bias_db': 0.0,
        'ldr_v_bias_db': 0.0,
        'minute': 49,
        'month': 4,
        'noise_h_co_dbm': -73.5886001586914,
        'noise_h_cx_dbm': -73.5886001586914,
        'noise_source_h_dbm': -9999.0,
        'noise_source_v_dbm': -9999.0,
        'noise_v_co_dbm': -73.13050079345703,
        'noise_v_cx_dbm': -73.13050079345703,
        'power_meas_loss_h_db': 4.96999979019165,
        'power_meas_loss_v_db': 5.840000152587891,
        'pulse_width_us': 0.800000011920929,
        'radar_constant_h_db': 73.25920104980469,
        'radar_constant_v_db': 73.25920104980469,
        'radar_name': 'C-SAPR',
        'rx_gain_h_co_dbm': 33.31079864501953,
        'rx_gain_h_cx_dbm': 33.31079864501953,
        'rx_gain_v_co_dbm': 34.24689865112305,
        'rx_gain_v_cx_dbm': 34.24689865112305,
        'rx_slope_h_co_db': 1.0030399560928345,
        'rx_slope_h_cx_db': 1.0030399560928345,
        'rx_slope_v_co_db': 0.9963300228118896,
        'rx_slope_v_cx_db': 0.9963300228118896,
        'second': 36,
        'spare': (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                  0.0, 0.0),
        'sun_h_co_dbm': -9999.0,
        'sun_h_cx_dbm': -9999.0,
        'sun_v_co_dbm': -9999.0,
        'sun_v_cx_dbm': -9999.0,
        'system_phidp_deg': -60.0,
        'test_pulse_h_dbm': -9999.0,
        'test_pulse_v_dbm': -9999.0,
        'twoway_radome_loss_h_db': 0.5,
        'twoway_radome_loss_v_db': 0.5,
        'twoway_waveguide_loss_h_db': 4.0,
        'twoway_waveguide_loss_v_db': 4.0,
        'wavelength_cm': 5.329639911651611,
        'xmit_power_h_dbm': 83.6500015258789,
        'xmit_power_v_dbm': 83.6500015258789,
        'year': 2011,
        'zdr_bias_db': -1.2799999713897705,
        'zh1km_co_dbz': -33.63949966430664,
        'zh1km_cx_dbz': -33.63949966430664,
        'zv1km_co_dbz': -34.11470031738281,
        'zv1km_cx_dbz': -34.11470031738281}

    for k, v in ref_calib_info.items():
        print("checking key:", k)
        assert mdvfile.calib_info[k] == v


def test_radar_info():

    # check radar_info
    ref_radar_info = {
        'altitude_km': 0.32760000228881836,
        'antenna_gain_db': 44.5,
        'field_flag': 0,
        'follow_mode': 1,
        'gate_spacing_km': 0.11991698294878006,
        'horiz_beam_width_deg': 1.0,
        'latitude_deg': 36.79615783691406,
        'longitude_deg': -97.45054626464844,
        'measXmitPowerDbmH_dbm': 82.7537612915039,
        'measXmitPowerDbmV_dbm': 82.65692138671875,
        'nfields': 1,
        'nfields_current': 0,
        'ngates': 110,
        'polarization': 5,
        'prf_hz': 1239.9993896484375,
        'prf_mode': 1,
        'prt2_s': 0.0,
        'prt_s': 0.0008064520079642534,
        'pulse_width_us': 0.800000011920929,
        'radar_constant': 73.25920104980469,
        'radar_id': 0,
        'radar_name': 'C-SAPR',
        'radar_type': 0,
        'receiver_gain_db': 33.31079864501953,
        'receiver_mds_dbm': -106.89939880371094,
        'samples_per_beam': 128,
        'scan_mode': 8,
        'scan_type': 0,
        'scan_type_name': 'Default_s',
        'spare_floats': (0.0, 0.0, 0.0, 0.0),
        'spare_ints': (0, 0),
        'start_range_km': 0.11787839233875275,
        'system_gain_db': 77.81079864501953,
        'unambig_range_km': 117.99627685546875,
        'unambig_vel_mps': 16.52467918395996,
        'vert_beam_width_deg': 1.0,
        'wavelength_cm': 5.330544471740723,
        'xmit_peak_pwr_watts': 231739552.0}

    for k, v in ref_radar_info.items():
        print("checking key:", k)
        assert mdvfile.radar_info[k] == v


def test_geometry():
    # test geometry attributes
    az_deg, range_km, el_deg = mdvfile._calc_geometry()
    assert np.all(az_deg == np.arange(360))
    assert len(range_km) == 110
    assert_almost_equal(range_km[10], 1.32, 2)
    assert len(el_deg) == 1
    assert el_deg[0] == 0.75


def test_geometry_raises():
    mdvfile = MdvFile(pyart.testing.MDV_GRID_FILE)
    pytest.raises(NotImplementedError, mdvfile._calc_geometry)


def test_mdv_time():
    # test times attribute
    assert mdvfile.times['time_begin'] == datetime(2011, 5, 20, 11, 1)
    assert mdvfile.times['time_centroid'] == datetime(2011, 5, 20, 11, 6, 35)
    assert mdvfile.times['time_end'] == datetime(2011, 5, 20, 11, 6, 35)


def test_cart():
    # test cartography information
    assert mdvfile.projection == 'ppi'
    carts = mdvfile._make_carts_dict()
    assert carts['x'].shape == (1, 360, 110)
    assert carts['y'].shape == (1, 360, 110)
    assert carts['z'].shape == (1, 360, 110)
    assert_almost_equal(carts['x'][0, 1, 2], 6.24, 2)
    assert_almost_equal(carts['y'][0, 1, 2], 357.63, 2)
    assert_almost_equal(carts['z'][0, 1, 2], 4.69, 2)


def test_fields():
    # test fields
    ref_fields = ['DBZ_F', ]
    for a, b in zip(ref_fields, mdvfile.fields):
        assert a == b


def test_fileptr():
    # test fileptr
    assert hasattr(mdvfile.fileptr, 'read')


def test_read_one_field():

    mdvfile = MdvFile(pyart.testing.MDV_PPI_FILE)
    # extract a field
    assert mdvfile.fields_data[0] is None
    sweeps = mdvfile.read_a_field(0)
    assert sweeps.shape == (1, 360, 110)
    assert_almost_equal(sweeps[0, 1, 2], 13.19, 2)
    assert mdvfile.fields_data[0] is not None

    # reread, should be faster
    sweeps = mdvfile.read_a_field(0)
    assert sweeps.shape == (1, 360, 110)
    assert_almost_equal(sweeps[0, 1, 2], 13.19, 2)
    assert mdvfile.fields_data[0] is not None


def test_read_all_fields():

    mdvfile = MdvFile(pyart.testing.MDV_PPI_FILE)
    # read all fields
    assert mdvfile.fields_data[0] is None
    mdvfile.read_all_fields()
    assert mdvfile.fields_data[0] is not None
    assert_almost_equal(mdvfile.fields_data[0][0, 1, 2], 13.19, 2)


def test_read_all_fields_on_creation():
    mdvfile2 = MdvFile(pyart.testing.MDV_PPI_FILE, read_fields=True)
    assert mdvfile2.fields_data[0] is not None
    assert_almost_equal(mdvfile2.fields_data[0][0, 1, 2], 13.19, 2)
    mdvfile2.close()


def test_mdvfile_radar_stubs():
    mdvfile = pyart.io.mdv_common.MdvFile(None)
    # These methods are included in MdvFile as a stub for a future
    # write_radar_mdv function, test that the dictionaries they return have the
    # correct length.
    dic = mdvfile._get_chunk_header()
    assert len(dic) == 8
    assert isinstance(dic, dict)

    dic = mdvfile._get_radar_info()
    assert len(dic) == 38
    assert isinstance(dic, dict)

    dic = mdvfile._get_calib()
    assert len(dic) == 59
    assert isinstance(dic, dict)

    dic = mdvfile._get_compression_info()
    assert len(dic) == 5
    assert isinstance(dic, dict)

    dic = mdvfile._get_levels_info(10)
    assert len(dic) == 2
    assert isinstance(dic, dict)


def test_rhi_cart():
    mdvfile = pyart.io.mdv_common.MdvFile(pyart.testing.MDV_RHI_FILE)
    assert mdvfile.projection == 'rhi'
    carts = mdvfile._make_carts_dict()
    assert carts['x'].shape == (1, 283, 125)
    assert carts['y'].shape == (1, 283, 125)
    assert carts['z'].shape == (1, 283, 125)
    assert_almost_equal(carts['x'][0, 1, 2], -52.63, 2)
    assert_almost_equal(carts['y'][0, 1, 2], -332.31, 2)
    assert_almost_equal(carts['z'][0, 1, 2], 121.47, 2)


class Mdv_common_Tests(object):
    """
    Class for declaring unit tests for the io/mdv_common.py module which
    require a temporary file which will be removed at the end of the test.
    """

    def test_mdv_file_read_write_radar(self):
        with warnings.catch_warnings(record=True) as w:
            mdvfile_orig = pyart.io.mdv_common.MdvFile(
                pyart.testing.MDV_PPI_FILE)
            mdvfile_orig.read_all_fields()

            inmemfile = BytesIO()
            mdvfile_orig.write(inmemfile)
            inmemfile.seek(0)
            # check that a UserWarning was issued since zlib compression used
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)

            mdvfile = pyart.io.mdv_common.MdvFile(inmemfile)
            self.check_mdvfile_ppi(mdvfile)

    def test_read_write_file_objects(self):
        # read from and write two a file object
        f = open(pyart.testing.MDV_PPI_FILE, 'rb')
        mdvfile = pyart.io.mdv_common.MdvFile(f)
        mdvfile.read_all_fields()
        self.check_mdvfile_ppi(mdvfile)

        # write out the file using an file handler
        f2 = BytesIO()
        with warnings.catch_warnings():
            # ignore the UserWarning about the non-implement compression_type
            warnings.simplefilter("ignore", category=UserWarning)
            mdvfile.write(f2)
        f.close()
        f2.seek(0)

        # re-read and check object
        mdvfile = pyart.io.mdv_common.MdvFile(f2)
        self.check_mdvfile_ppi(mdvfile)

    @staticmethod
    def check_mdvfile_ppi(mdvfile):
        # check some parameters
        assert mdvfile.master_header['time_end'] == 1305889595
        assert mdvfile.master_header['chunk_hdr_offset'] == 2464
        assert len(mdvfile.field_headers) == 1
        assert mdvfile.field_headers[0]['bad_data_value'] == 0.0
        assert len(mdvfile.vlevel_headers) == 1
        assert mdvfile.vlevel_headers[0]['record_len1'] == 1016
        assert mdvfile.chunk_headers[0]['chunk_id'] == 3
        assert mdvfile.calib_info['day'] == 15
        assert mdvfile.radar_info['scan_mode'] == 8
        az_deg, range_km, el_deg = mdvfile._calc_geometry()
        assert np.all(az_deg == np.arange(360))
        assert len(range_km) == 110
        assert_almost_equal(range_km[10], 1.32, 2)
        assert len(el_deg) == 1
        assert el_deg[0] == 0.75
        assert mdvfile.times['time_begin'] == datetime(2011, 5, 20, 11, 1)
        assert len(mdvfile.elevations) == 0

    def test_mdv_file_read_write_radar_rhi(self):
        # write and read the RHI file which contains a elevation chunk
        mdvfile = pyart.io.mdv_common.MdvFile(pyart.testing.MDV_RHI_FILE)
        mdvfile.read_all_fields()
        inmemfile = BytesIO()
        with warnings.catch_warnings():
            # ignore the UserWarning about the non-implement compression_type
            warnings.simplefilter("ignore", category=UserWarning)
            mdvfile.write(inmemfile)
        inmemfile.seek(0)
        mdvfile2 = pyart.io.mdv_common.MdvFile(inmemfile)

        # verify that the elevations are similar
        assert len(mdvfile2.elevations) == len(mdvfile.elevations)
        assert mdvfile2.elevations[0] == mdvfile.elevations[0]



"""
Estimation of noise in Doppler spectra using the Hildebrand Sekhon method.

"""

import numpy as np

def estimate_noise_hs74(spectrum, navg=1, nnoise_min=1):
    """
    Estimate noise parameters of a Doppler spectrum.

    Use the method of estimating the noise level in Doppler spectra outlined
    by Hildebrand and Sehkon, 1974.

    Parameters
    ----------
    spectrum : array like
        Doppler spectrum in linear units.
    navg : int, optional
        The number of spectral bins over which a moving average has been
        taken. Corresponds to the **p** variable from equation 9 of the
        article. The default value of 1 is appropriate when no moving
        average has been applied to the spectrum.
    nnoise_min : int, optional
        Minimum number of noise samples to consider the estimation valid.

    Returns
    -------
    mean : float-like
        Mean of points in the spectrum identified as noise.
    threshold : float-like
        Threshold separating noise from signal. The point in the spectrum with
        this value or below should be considered as noise, above this value
        signal. It is possible that all points in the spectrum are identified
        as noise. If a peak is required for moment calculation then the point
        with this value should be considered as signal.
    var : float-like
        Variance of the points in the spectrum identified as noise.
    nnoise : int
        Number of noise points in the spectrum.

    References
    ----------
    P. H. Hildebrand and R. S. Sekhon, Objective Determination of the Noise
    Level in Doppler Spectra. Journal of Applied Meteorology, 1974, 13,
    808-811.

    """
    sorted_spectrum = np.sort(spectrum)
    nnoise = len(spectrum)  # default to all points in the spectrum as noise

    rtest = 1+1/navg
    sum1 = 0.
    sum2 = 0.
    for i, pwr in enumerate(sorted_spectrum):
        npts = i+1
        sum1 += pwr
        sum2 += pwr*pwr

        if npts < nnoise_min:
            continue

        if npts*sum2 < sum1*sum1*rtest:
            nnoise = npts
        else:
            # partial spectrum no longer has characteristics of white noise.
            sum1 -= pwr
            sum2 -= pwr*pwr
            break

    mean = sum1/nnoise
    var = sum2/nnoise-mean*mean
    threshold = sorted_spectrum[nnoise-1]
    return mean, threshold, var, nnoise





def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('pyart', parent_package, top_path)
    config.add_subpackage('io')     # io first to detect if RSL is missing.
    config.add_subpackage('__check_build')
    config.add_subpackage('core')
    config.add_subpackage('correct')
    config.add_subpackage('graph')
    config.add_subpackage('map')
    config.add_subpackage('retrieve')
    config.add_subpackage('filters')
    config.add_subpackage('testing')
    config.add_subpackage('util')
    config.add_subpackage('aux_io')
    config.add_subpackage('bridge')

    config.add_data_dir('tests')
    return config

if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())
